# base imports
import logging
import pims
import cv2
from sklearn.cluster import DBSCAN
from collections import Counter
import pandas as pd
from pathlib import Path
from tqdm import tqdm
import numpy as np
from PIL import Image as PILImage, ImageDraw

# Logging
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


def drawBoxes(df: pd.DataFrame, movie_dir: str, out_path: str):
    """
    For each unique movie, create a dictionary of movie paths and their corresponding pims.Video
    objects. Then, for each unique movie, frame number, species id, and filename, get the corresponding
    frame, get the bounding boxes for that frame, and draw the bounding boxes on the frame. Then, write
    the frame to the output directory

    :param df: the dataframe containing the bounding box coordinates
    :param movie_dir: The directory where the movies are stored
    :param out_path: The path to the directory where you want to save the images with the bounding boxes
           drawn on them
    :return:
    """
    df["movie_path"] = df["filename"].apply(
        lambda x: str(
            (Path(movie_dir) / Path(x).name.rsplit("_frame_")[0]).with_suffix(".mp4")
        )
    )
    movie_dict = {i: pims.Video(i) for i in df["movie_path"].unique()}
    df["annotation"] = df[["x_position", "y_position", "width", "height"]].apply(
        lambda x: tuple([x[0], x[1], x[2], x[3]]), 1
    )
    df = df.drop(columns=["x_position", "y_position", "width", "height"])
    for name, group in tqdm(
        df.groupby(["movie_path", "frame_number", "species_id", "filename"])
    ):
        frame = movie_dict[name[0]][name[1]]
        boxes = [tuple(i[4:])[0] for i in group.values]
        for box in boxes:
            # Calculating end-point of bounding box based on starting point and w, h
            end_box = tuple([int(box[0] + box[2]), int(box[1] + box[3])])
            # changed color and width to make it visible
            cv2.rectangle(frame, (int(box[0]), int(box[1])), end_box, (255, 0, 0), 1)
        out_dir = Path(out_path)
        out_dir.mkdir(parents=True, exist_ok=True)
        # Recursively add permissions to folders created
        for root, dirs, files in out_dir.iterdir():
            Path(root).chmod(0o777)
        cv2.imwrite(out_dir / Path(name[3]).name, frame)


def bb_iou(boxA, boxB):
    """
    The function takes two bounding boxes, computes the area of intersection, and divides it by the area
    of the union of the two boxes

    :param boxA: The first bounding box
    :param boxB: The ground truth box
    :return: The IOU value
    """

    # Compute edges
    temp_boxA = boxA.copy()
    temp_boxB = boxB.copy()
    temp_boxA[2], temp_boxA[3] = (
        temp_boxA[0] + temp_boxA[2],
        temp_boxA[1] + temp_boxA[3],
    )
    temp_boxB[2], temp_boxB[3] = (
        temp_boxB[0] + temp_boxB[2],
        temp_boxB[1] + temp_boxB[3],
    )

    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(temp_boxA[0], temp_boxB[0])
    yA = max(temp_boxA[1], temp_boxB[1])
    xB = min(temp_boxA[2], temp_boxB[2])
    yB = min(temp_boxA[3], temp_boxB[3])

    # compute the area of intersection rectangle
    interArea = abs(max((xB - xA, 0)) * max((yB - yA), 0))
    if interArea == 0:
        return 1
    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = abs((temp_boxA[2] - temp_boxA[0]) * (temp_boxA[3] - temp_boxA[1]))
    boxBArea = abs((temp_boxB[2] - temp_boxB[0]) * (temp_boxB[3] - temp_boxB[1]))

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the intersection area
    iou = interArea / float(boxAArea + boxBArea - interArea)

    # return the intersection over union value
    return 1 - iou


def filter_bboxes(
    total_users: int, users: list, bboxes: list, obj: float, eps: float, iua: float
):
    """
    If at least half of the users who saw this frame decided that there was an object, then we cluster
    the bounding boxes based on the IoU criterion. If at least 80% of users agree on the annotation,
    then we accept the cluster assignment

    :param total_users: total number of users who saw this frame
    :param users: list of user ids
    :param bboxes: list of bounding boxes
    :param obj: the minimum fraction of users who must have seen an object in order for it to be considered
    :param eps: The maximum distance between two samples for them to be considered as in the same neighborhood
    :param iua: the minimum percentage of users who must agree on a bounding box for it to be accepted
    """

    # If at least half of those who saw this frame decided that there was an object
    user_count = pd.Series(users).nunique()
    if user_count / total_users >= obj:
        # Get clusters of annotation boxes based on iou criterion
        cluster_ids = DBSCAN(min_samples=1, metric=bb_iou, eps=eps).fit_predict(bboxes)
        # Count the number of users within each cluster
        counter_dict = Counter(cluster_ids)
        # Accept a cluster assignment if at least 80% of users agree on annotation
        passing_ids = [k for k, v in counter_dict.items() if v / user_count >= iua]

        indices = np.isin(cluster_ids, passing_ids)

        final_boxes = []
        for i in passing_ids:
            # Compute median over all accepted bounding boxes
            boxes = np.median(np.array(bboxes)[np.where(cluster_ids == i)], axis=0)
            final_boxes.append(boxes)

        return indices, final_boxes

    else:
        return [], bboxes


def draw_annotations_in_frame(im: PILImage.Image, class_df_subject: pd.DataFrame):
    """
    > The function takes an image and a dataframe of annotations and returns the image with the
    annotations drawn on it

    :param im: the image object of type PILImage
    :param class_df_subject: a dataframe containing the annotations for a single subject
    :return: The image with the annotations
    """
    # Calculate image size
    dw, dh = im._size

    # Draw rectangles of each annotation
    img1 = ImageDraw.Draw(im)

    # Merge annotation info into a tuple
    class_df_subject["vals"] = class_df_subject[["x", "y", "w", "h"]].values.tolist()

    for index, row in class_df_subject.iterrows():
        # Specify the vals object
        vals = row.vals

        # Adjust annotantions to image size
        vals_adjusted = tuple(
            [
                int(vals[0]),
                int(vals[1]),
                int((vals[0] + vals[2])),
                int((vals[1] + vals[3])),
            ]
        )

        # Draw annotation
        img1.rectangle(vals_adjusted, width=2)

    return im



import hashlib
import itertools
import os
from collections import OrderedDict
from dataclasses import dataclass
from typing import Optional, TypeVar, Generic, Hashable, Any
from typing import Tuple, Iterator

import av
import cv2
from more_itertools import first
import numpy as np
import sys

KeyType = TypeVar("KeyType")
ItemType = TypeVar("ItemType")
BGRImageArray = np.ndarray
TimeIntervalTuple = Tuple[Optional[float], Optional[float]]


def get_memory_footprint(item: Any) -> int:
    # TODO: Recurse through dataclasses
    if isinstance(item, np.ndarray):
        return item.itemsize * item.size
    else:
        return sys.getsizeof(item)  # Only returns pointer-size, no recursion


class CacheDict(Generic[KeyType, ItemType]):
    """A simple buffer that just keeps a cache of recent entries
    Example
        cache = ItemCache(buffer_len=3)
        cache[1]='aaa'
        cache[5]='bbb'
        cache[7]='ccc'
    """

    def __init__(
        self,
        buffer_length: Optional[int] = None,
        buffer_size_bytes: Optional[int] = None,
        calculate_size_once=True,
        always_allow_one_item: bool = False,
    ):
        self._buffer = OrderedDict()
        self._buffer_length = buffer_length
        self._buffer_size_bytes = buffer_size_bytes
        self._first_object_size: Optional[int] = None
        self._calculate_size_once = calculate_size_once
        self._current_buffer_size = 0
        self._always_allow_one_item = always_allow_one_item

    def _remove_oldest_item(self):
        if len(self._buffer) > 0:
            first_key = first(self._buffer.keys())
            value, itemsize = self._buffer[first_key]
            del self._buffer[first(self._buffer.keys())]
            if itemsize is not None:
                self._current_buffer_size -= itemsize

    def __setitem__(self, key: Hashable, value: ItemType) -> None:
        size = None
        if self._buffer_length is not None and len(self._buffer) == self._buffer_length:
            self._remove_oldest_item()

        if self._buffer_size_bytes is not None:
            size = (
                get_memory_footprint(value)
                if not self._calculate_size_once or self._first_object_size is None
                else self._first_object_size
            )
            while (
                len(self._buffer) > 0
                and self._current_buffer_size + size > self._buffer_size_bytes
            ):
                self._remove_oldest_item()

            self._current_buffer_size += size

            if (
                size < self._buffer_size_bytes
                or len(self._buffer) == 0
                and self._always_allow_one_item
            ):
                self._buffer[key] = value, size
        else:
            self._buffer[key] = value, size

    def __getitem__(self, key: Hashable) -> ItemType:
        if key in self._buffer:
            value, _ = self._buffer[key]
            return value
        else:
            raise KeyError(f"{key} is not in cache")

    def __contains__(self, key: Hashable):
        return key in self._buffer


def fit_image_to_max_size(image: BGRImageArray, max_size: Tuple[int, int]):
    """Make sure image fits within (width, height) max_size while preserving aspect ratio"""
    if image.shape[0] > max_size[1] or image.shape[1] > max_size[0]:
        scale_factor = min(max_size[1] / image.shape[0], max_size[0] / image.shape[1])
        return cv2.resize(src=image, dsize=None, fx=scale_factor, fy=scale_factor)
    else:
        return image


def compute_fixed_hash(array):
    return hashlib.md5(array.tostring()).hexdigest()


@dataclass
class VideoFrameInfo:
    image: BGRImageArray
    seconds_into_video: float
    frame_ix: int
    fps: float

    def get_size_xy(self) -> Tuple[int, int]:
        return self.image.shape[1], self.image.shape[0]

    def get_progress_string(self) -> str:
        return f"t={self.seconds_into_video:.2f}s, frame={self.frame_ix}"


@dataclass
class VideoReader:
    """
    The reader efficiently provides access to video frames.
    It uses pyav: https://pyav.org/docs/stable/
    Usage:
        reader = VideoReader(path=video_segment.path, use_cache=use_cache)
        # Iterate in order
        for frame in reader.iter_frames(time_interval=(1, 2)):
            cv2.imshow('frame', frame.image)
            cv2.waitKey(1)
        # Request individually
        frame = reader.request_frame(20)  # Ask for the 20th frame
        cv2.imshow('frame', frame.image)
        cv2.waitKey(1)
    Implementation:
    - Providing them in order should be FAST
    - Requesting the same frame twice or backtracking a few frames should be VERY FAST (ie - use a cache)
    - Requesting random frames should be reasonably fast (do not scan from start)
    Note: Due to bug in OpenCV with GET_PROP_POS_FRAMES
        https://github.com/opencv/opencv/issues/9053
        We use "av": conda install av -c conda-forge
    """

    def __init__(
        self,
        path: str,
        buffer_size_bytes=1024**3,
        threshold_frames_to_scan=30,
        max_size_xy: Optional[Tuple[int, int]] = None,
        use_cache: bool = True,
    ):
        self._path = os.path.expanduser(path)
        assert os.path.exists(self._path), f"Cannot find a video at {path}"

        self.container = av.container.open(self._path)
        # self.stream = self.container.streams.video[0]

        # self._cap = cv2.VideoCapture(path)
        self._frame_cache: CacheDict[int, VideoFrameInfo] = CacheDict(
            buffer_size_bytes=buffer_size_bytes, always_allow_one_item=True
        )
        self._next_index_to_be_read: int = 0
        self._threshold_frames_to_scan = threshold_frames_to_scan
        # self._fps = self._cap.get(cv2.CAP_PROP_FPS)
        self._fps = float(self.container.streams.video[0].guessed_rate)

        self._n_frames = self.container.streams.video[0].frames
        self._cached_last_frame: Optional[VideoFrameInfo] = (
            None  # Helps fix weird bug... see notes below
        )
        self._max_size_xy = max_size_xy
        self._use_cache = use_cache
        self._iterator = self._iter_frame_data()

    def get_n_frames(self) -> int:
        return self._n_frames

    def time_to_nearest_frame(self, t: float) -> int:
        return max(0, min(self.get_n_frames() - 1, round(t * self._fps)))

    def frame_index_to_nearest_frame(self, index: int) -> int:
        return max(0, min(self.get_n_frames() - 1, index))

    def iter_frame_ixs(
        self,
        time_interval: TimeIntervalTuple = (None, None),
        frame_interval: Tuple[Optional[int], Optional[int]] = (None, None),
    ) -> Iterator[int]:
        assert time_interval == (None, None) or frame_interval == (
            None,
            None,
        ), "You can provide a time interval or frame inteval, not both"
        if time_interval != (None, None):
            tstart, tstop = time_interval
            return range(
                self.time_to_nearest_frame(tstart) if tstart is not None else 0,
                (
                    self.time_to_nearest_frame(tstop) + 1
                    if tstop is not None
                    else self.get_n_frames()
                ),
            )
        elif frame_interval != (None, None):
            istart, istop = frame_interval
            return range(
                self.frame_index_to_nearest_frame(istart) if istart is not None else 0,
                (
                    self.frame_index_to_nearest_frame(istop)
                    if istop is not None
                    else self.get_n_frames()
                ),
            )
        else:
            return range(0, self.get_n_frames())

    def iter_frames(
        self,
        time_interval: TimeIntervalTuple = (None, None),
        frame_interval: Tuple[Optional[int], Optional[int]] = (None, None),
    ) -> Iterator[VideoFrameInfo]:
        for i in self.iter_frame_ixs(
            time_interval=time_interval, frame_interval=frame_interval
        ):
            yield self.request_frame(i)

    def _iter_frame_data(self):
        for frame in self.container.decode(self.container.streams.video[0]):
            yield frame

    def request_frame(self, index: int) -> VideoFrameInfo:
        """
        Request a frame of the video.  If the requested frame is out of bounds, this will return the frame
        on the closest edge.
        """
        if index < 0:
            index = self.get_n_frames() + index
        index = max(0, min(self.get_n_frames() - 1, index))

        if index == self.get_n_frames() - 1 and self._cached_last_frame is not None:
            return (
                self._cached_last_frame
            )  # There's a weird bug preventing us from loading the last frame again
        elif index in self._frame_cache:
            return self._frame_cache[index]
        elif 0 <= index - self._next_index_to_be_read < self._threshold_frames_to_scan:
            frame = None
            for _ in range(self._next_index_to_be_read, index + 1):
                try:
                    frame_data = next(self._iterator)
                except StopIteration:
                    raise Exception(
                        f"Could not get frame at index {index}, despite n_frames being {self.get_n_frames()}"
                    )

                image = frame_data.to_rgb().to_ndarray(format="bgr24")

                if self._max_size_xy is not None:
                    image = fit_image_to_max_size(image, self._max_size_xy)
                frame = VideoFrameInfo(
                    image=image,
                    seconds_into_video=self._next_index_to_be_read / self._fps,
                    frame_ix=self._next_index_to_be_read,
                    fps=self._fps,
                )
                if self._use_cache:
                    self._frame_cache[frame.frame_ix] = frame
                self._next_index_to_be_read += 1

            assert frame is not None, f"Error loading video frame at index {index}"
            return frame
        else:
            max_seek_search = 100
            stream = self.container.streams.video[0]
            pts = int(index * stream.duration / stream.frames)
            self.container.seek(pts, stream=stream)
            self._iterator = self._iter_frame_data()
            for j, f in enumerate(self._iterator):
                if j > max_seek_search:
                    raise RuntimeError(
                        f"Did not find target within {max_seek_search} frames of seek"
                    )
                if f.pts >= pts - 1:
                    self._iterator = itertools.chain([f], self._iterator)
                    break
            self._next_index_to_be_read = index
            return self.request_frame(index)


def test_video_reader(video_path: str, show=False):
    for use_cache in (False, True):
        reader = VideoReader(path=video_path, use_cache=use_cache)

        all_frame_hashes = [compute_fixed_hash(f.image) for f in reader.iter_frames()]
        all_frame_hashes_again = [
            compute_fixed_hash(f.image) for f in reader.iter_frames()
        ]
        assert all_frame_hashes == all_frame_hashes_again

        # With just the top lines:
        # Starting block 'Running with use_cache=False...
        #   EasyProfile: Running with use_cache=False took 2260.51ms
        # Starting block 'Running with use_cache=True...
        # EasyProfile: Running with use_cache=True took 1942.80ms
        count = 0
        time_snoppet_hashes = []
        for frame in reader.iter_frames(time_interval=(1, 2)):
            if show:
                cv2.imshow("frame", frame.image)
                cv2.waitKey(1)
            time_snoppet_hashes.append(compute_fixed_hash(frame.image))
            count += 1
        start_ix = reader.time_to_nearest_frame(1)
        assert count > 20
        assert all_frame_hashes[start_ix : start_ix + count] == time_snoppet_hashes

        index_snippet_hashes = []

        for i, frame in enumerate(reader.iter_frames(frame_interval=(10, 20))):
            if show:
                cv2.imshow("frame", frame.image)
                cv2.waitKey(1)
            index_snippet_hashes.append(compute_fixed_hash(frame.image))
            assert frame.frame_ix == 10 + i
        assert len(index_snippet_hashes) == 10
        assert all_frame_hashes[10:20] == index_snippet_hashes

        last_frame = reader.request_frame(-1)
        assert last_frame.frame_ix == reader.get_n_frames() - 1
        assert compute_fixed_hash(last_frame.image) == all_frame_hashes[-1]
        assert (
            reader.request_frame(reader.get_n_frames() - 1).frame_ix
            == reader.get_n_frames() - 1
        )
        assert (
            reader.request_frame(reader.get_n_frames()).frame_ix
            == reader.get_n_frames() - 1
        )

        second_last_frame = reader.request_frame(reader.get_n_frames() - 2)
        assert second_last_frame.frame_ix == reader.get_n_frames() - 2
        assert compute_fixed_hash(second_last_frame.image) == all_frame_hashes[-2]

        third_last_frame = reader.request_frame(reader.get_n_frames() - 3)
        assert third_last_frame.frame_ix == reader.get_n_frames() - 3
        assert compute_fixed_hash(third_last_frame.image) == all_frame_hashes[-3]

        last_frame_again = reader.request_frame(-1)
        assert compute_fixed_hash(last_frame_again.image) == all_frame_hashes[-1]

        first_frame = reader.request_frame(0)
        assert first_frame.frame_ix == 0
        assert compute_fixed_hash(first_frame.image) == all_frame_hashes[0]

        random_frame = reader.request_frame(20)
        assert random_frame.frame_ix == 20
        assert compute_fixed_hash(random_frame.image) == all_frame_hashes[20]



# -*- coding: utf-8 -*-
# base imports
import logging
import sqlite3
import ftfy
import pandas as pd
from pathlib import Path

# Logging
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


def fix_text_encoding(string: str):
    """This function corrects for text encoding errors, which occur when there is
    for example an ä,å,ö present."""
    return ftfy.fix_text(string)


def fix_text_encoding_folder(folder_name):
    """
    This function corrects for text encoding errors, which occur when there is
    for example an ä,å,ö present. It runs through all the file and folder names
    of the directory you give it. It uses the package ftfy, which recognizes
    which encoding the text has based on the text itself, and it encodes/decodes
    it to utf8.
    This function was tested on a Linux and Windows device with package version
    6.1.1. With package version 5.8 it did not work.

    This function can replace the unswedify and reswedify functions from
    koster_utils, but this is not implemented yet.
    """
    for item in Path(folder_name).iterdir():
        if item.is_dir():
            for sub_item in item.iterdir():
                old_path = sub_item
                new_path = sub_item.parent / ftfy.fix_text(sub_item.name)
                old_path.rename(new_path)


def get_koster_col_names(table_name: str):
    """Return a dictionary with the project-specific column names of a csv of interest
    This function helps matching the schema format without modifying the column names of the original csv.

    :param table_name: a string of the name of the schema table of interest
    :return: a dictionary with the names of the columns
    """

    if table_name == "movies":
        # Save the column names of interest in a dict
        col_names_dic = {
            "SamplingStart": "sampling_start",
            "SamplingEnd": "sampling_end",
        }
    else:
        # Create empty data frame as there are no project-specific
        # columns for this table
        col_names_dic = {}

    return col_names_dic


# Function to process subjects uploaded automatically
def auto_subjects(subjects_df: pd.DataFrame, auto_date: str):
    """
    The function `auto_subjects` selects and extracts metadata from subjects that were automatically
    uploaded after a specified date.

    :param subjects_df: `subjects_df` is a pandas DataFrame containing information about subjects, such
    as their IDs, project IDs, and creation dates
    :type subjects_df: pd.DataFrame
    :param auto_date: auto_date is a string parameter that represents the date from which the function
    should select automatically uploaded frames. The function will only select frames that were created
    after this date
    :type auto_date: str
    :return: The function `auto_subjects` returns a pandas DataFrame containing metadata information
    extracted from subjects that were automatically uploaded after a specified date.
    """
    # Select automatically uploaded frames
    auto_subjects_df = subjects_df[(subjects_df["created_at"] > auto_date)]

    from kso_utils.zooniverse_utils import extract_metadata

    # Extract metadata from automatically uploaded subjects
    auto_subjects_df, auto_subjects_meta = extract_metadata(auto_subjects_df)

    # Combine metadata info with the subjects df
    auto_subjects_df = pd.concat([auto_subjects_df, auto_subjects_meta], axis=1)

    return auto_subjects_df


# Function to process subjects uploaded manually
def manual_subjects(subjects_df: pd.DataFrame, manual_date: str, auto_date: str):
    """
    The function extracts metadata from manually uploaded clips and processes it to combine with the
    subjects dataframe.

    :param subjects_df: A pandas DataFrame containing information about subjects, including metadata and
    creation dates
    :type subjects_df: pd.DataFrame
    :param manual_date: The date from which to start selecting clips uploaded manually
    :type manual_date: str
    :param auto_date: It seems like the parameter auto_date is missing from the code snippet. Can you
    provide more information on what this parameter represents?
    :type auto_date: str
    :return: a pandas DataFrame containing information about clips that were uploaded manually, along
    with their metadata and processed information.
    """
    from kso_utils.zooniverse_utils import extract_metadata

    # Select clips uploaded manually
    man_clips_df = (
        subjects_df[
            (subjects_df["metadata"].str.contains(".mp4"))
            & (subjects_df["created_at"].between(manual_date, auto_date))
        ]
        .reset_index(drop=True)
        .reset_index()
    )

    # Specify the type of subject
    man_clips_df["subject_type"] = "clip"

    # Extract metadata from manually uploaded clips
    man_clips_df, man_clips_meta = extract_metadata(man_clips_df)

    if len(man_clips_meta) > 0:
        # Process the metadata of manually uploaded clips
        man_clips_meta = process_manual_clips(man_clips_meta)

        # Combine metadata info with the subjects df
        man_clips_df = pd.concat([man_clips_df, man_clips_meta], axis=1)

    return man_clips_df


# Function to process the metadata of clips that were uploaded manually
def process_manual_clips(meta_df: pd.DataFrame):
    """
    The function processes metadata of manual clips by extracting relevant information such as clip
    start and end times and the filename of the original movie.

    :param meta_df: The input parameter `meta_df` is a Pandas DataFrame containing metadata information
    about video clips. It is assumed that the DataFrame has a column named "filename" which contains the
    name of the video clip file in the format "original_movie_name_start_time.mp4". The function
    processes this information to extract
    :type meta_df: pd.DataFrame
    :return: a pandas DataFrame containing the filename of the clips, the filename of the original
    movie, the starting time of the clips in relation to the original movie, and the end time of the
    clips in relation to the original movie.
    """
    # Select the filename of the clips and remove extension type
    clip_filenames = meta_df["filename"].str.replace(".mp4", "", regex=True)

    # Get the starting time of clips in relation to the original movie
    # split the filename and select the last section
    meta_df["clip_start_time"] = clip_filenames.str.rsplit("_", 1).str[-1]

    # Extract the filename of the original movie
    meta_df["movie_filename"] = meta_df.apply(
        lambda x: x["filename"]
        .replace("_" + x["clip_start_time"], "")
        .replace(".mp4", ".mov"),
        axis=1,
    )

    # Get the end time of clips in relation to the original movie
    meta_df["clip_start_time"] = pd.to_numeric(
        meta_df["clip_start_time"], downcast="signed"
    )
    meta_df["clip_end_time"] = meta_df["clip_start_time"] + 10

    # Select only relevant columns
    meta_df = meta_df[
        ["filename", "movie_filename", "clip_start_time", "clip_end_time"]
    ]

    return meta_df


def get_movies_id(df: pd.DataFrame, conn: sqlite3.Connection):
    """
    This function retrieves movie IDs based on movie filenames from a database and merges them with a
    given DataFrame.

    :param df: A pandas DataFrame containing information about movie filenames and clip subjects
    :type df: pd.DataFrame
    :param conn: SQL connection object
    :return: a pandas DataFrame with the movie_ids added to the input DataFrame based on matching movie
    filenames with the movies table in a SQLite database. The function drops the movie_filename column
    before returning the DataFrame.
    """
    from kso_utils.db_utils import get_df_from_db_table

    # Query id and filenames from the movies table
    movies_df = get_df_from_db_table(conn, "movies")[["id", "filename"]]
    movies_df = movies_df.rename(
        columns={"id": "movie_id", "filename": "movie_filename"}
    )

    # Check all the movies have a unique ID
    df_unique = df.movie_filename.unique()
    movies_df_unique = movies_df.movie_filename.unique()
    diff_filenames = set(df_unique).difference(movies_df_unique)

    if diff_filenames:
        raise ValueError(
            f"There are clip subjects that don't have movie_id. The movie filenames are {diff_filenames}"
        )

    # Reference the manually uploaded subjects with the movies table
    df = pd.merge(df, movies_df, how="left", on="movie_filename")

    # Drop the movie_filename column
    df = df.drop(columns=["movie_filename"])

    return df


def process_koster_subjects(subjects: pd.DataFrame, conn: sqlite3.Connection):
    """
    This function takes in a dataframe of subjects and a path to the database and returns a dataframe of
    subjects with updated metadata

    :param subjects: the dataframe of subjects from the database
    :type subjects: pd.DataFrame
    :param conn: SQL connection object
    :return: A dataframe with all the subjects that have been uploaded to the database.
    """

    # Set the date when the metadata of subjects uploaded matches/doesn't match schema.py requirements
    # Specify the date when the metadata of subjects uploaded matches schema.py
    auto_date = "2020-05-29 00:00:00 UTC"

    # Specify the starting date when clips were manually uploaded
    manual_date = "2019-11-17 00:00:00 UTC"

    # Select automatically uploaded subjects
    auto_subjects_df = auto_subjects(subjects, auto_date=auto_date)

    # Select manually uploaded subjects
    manual_subjects_df = manual_subjects(
        subjects, manual_date=manual_date, auto_date=auto_date
    )

    if len(manual_subjects_df) > 0:
        # Include movie_ids to the metadata
        manual_subjects_df = get_movies_id(manual_subjects_df, conn=conn)

        # Combine all uploaded subjects
        subjects = pd.merge(manual_subjects_df, auto_subjects_df, how="outer")

    else:
        subjects = auto_subjects_df

    return subjects


def process_clips_koster(annotations, row_class_id: str, rows_list: list):
    """
    For each annotation, if the task is T4, then for each species annotated, flatten the relevant
    answers and save the species of choice, class and subject id

    :param annotations: the list of annotations for a given classification
    :param row_class_id: the classification id of the row
    :param rows_list: list
    :type rows_list: list
    :return: A list of dictionaries, each dictionary containing the classification id, the label, the first time seen and how many individuals were seen.
    """

    nothing_values = [
        "NOANIMALSPRESENT",
        "ICANTRECOGNISEANYTHING",
        "ISEENOTHING",
        "NOTHINGHERE",
    ]

    for ann_i in annotations:
        if ann_i["task"] == "T4":
            # Select each species annotated and flatten the relevant answers
            for value_i in ann_i["value"]:
                choice_i = {}
                # If choice = 'nothing here', set follow-up answers to blank
                if value_i["choice"] in nothing_values:
                    f_time = ""
                    inds = ""
                # If choice = species, flatten follow-up answers
                else:
                    answers = value_i["answers"]
                    f_time, inds = None, None
                    for k in answers.keys():
                        if "FIRSTTIME" in k:
                            f_time = answers[k].replace("S", "")
                        if "INDIVIDUAL" in k:
                            inds = answers[k]
                        elif "FIRSTTIME" not in k and "INDIVIDUAL" not in k:
                            f_time, inds = None, None

                # Save the species of choice, class and subject id
                choice_i.update(
                    {
                        "classification_id": row_class_id,
                        "label": value_i["choice"],
                        "first_seen": f_time,
                        "how_many": inds,
                    }
                )

                rows_list.append(choice_i)

    return rows_list


def process_koster_movies_csv(movies_df: pd.DataFrame):
    """
    It takes a dataframe of movies and returns a dataframe of movies with the following changes:

    - The filename is standardized
    - The filename is unswedified
    - The filename is renamed to fpath
    - The SamplingStart and SamplingEnd columns are renamed to sampling_start and sampling_end

    :param movies_df: the dataframe containing the movies
    :return: A dataframe with the columns:
        - filename
        - fpath
        - sampling_start
        - sampling_end
    """
    # Standarise the filename
    movies_df["filename"] = movies_df["filename"].str.normalize("NFD")

    # Ensure the filename has standard characters
    movies_df["filename"] = movies_df["filename"].apply(lambda x: fix_text_encoding(x))

    # TO DO Include server's path to the movie files
    movies_df["fpath"] = (
        movies_df["filename"].replace(".MP4", ".mp4").replace(".mov", ".mp4")
    )

    # Rename relevant fields
    movies_df = movies_df.rename(
        columns={
            "SamplingStart": "sampling_start",
            "SamplingEnd": "sampling_end",
        }
    )

    return movies_df



# base imports
import logging
import multiprocessing

# Logging
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


def import_model_modules(module_names):
    """
    This function imports specified modules and returns them as a dictionary.

    :param module_names: A list of strings representing the names of the modules to be imported. The
    order of the names in the list should correspond to the order of the modules in the returned
    dictionary. In this case, the expected order is ["train", "detect", "val"]
    :return: a dictionary containing the imported modules with keys "train", "detect", and "val". If a
    module fails to import, an error message is logged and that module is not included in the returned
    dictionary.
    """
    importlib = __import__("importlib")
    modules = {}
    for module_name, module_full in zip(["train", "detect", "val"], module_names):
        try:
            modules[module_name] = importlib.import_module(module_full)
        except ModuleNotFoundError:
            logging.error(f"Module {module_name} could not be imported.")
    return modules


def import_modules(module_names, utils: bool = True, models: bool = False):
    """
    The function imports specified modules and returns them as a dictionary.

    :param module_names: A list of strings representing the names of the modules to be imported
    :param utils: A boolean parameter that specifies whether the module names provided are for utility
    modules located in the "kso_utils" package or not. If True, the function will prepend "kso_utils."
    to each module name before attempting to import it. If False, the function will attempt to import
    the module directly, defaults to True
    :type utils: bool (optional)
    :param models: A boolean parameter that determines whether the imported modules are related to
    models or not. If it is set to True, the function will assume that the module names correspond to
    the three model presets: "train", "detect", and "val". If it is set to False, the function will
    import the, defaults to False
    :type models: bool (optional)
    :return: a dictionary of imported modules, where the keys are the names of the modules and the
    values are the imported module objects.
    """
    importlib = __import__("importlib")
    modules = {}
    model_presets = ["train", "detect", "val"]
    for i, module_name in enumerate(module_names):
        if utils:
            module_full = "kso_utils." + module_name
        else:
            module_full = module_name
        try:
            if models:
                module_name = model_presets[i]
            modules[module_name] = importlib.import_module(module_full)
        except ModuleNotFoundError:
            logging.error(f"Module {module_name} could not be imported.")
    return modules


def parallel_map(func, iterable, args=()):
    """
    The function `parallel_map` uses multiprocessing to apply a given function to each element of an
    iterable in parallel.

    :param func: The function to be applied to each element of the iterable
    :param iterable: The iterable is a sequence of elements that can be iterated over, such as a list,
    tuple, or range object. The function `func` will be applied to each element of the iterable in
    parallel using multiple processes
    :param args: args is a tuple of additional arguments that can be passed to the function being mapped
    in parallel. These arguments will be unpacked and passed to the function along with the
    corresponding element from the iterable. If no additional arguments are needed, the default value of
    an empty tuple can be used
    :return: The function `parallel_map` returns a list of results obtained by applying the function
    `func` to each element of the `iterable` in parallel using multiple processes. The `args` parameter
    is optional and can be used to pass additional arguments to the function `func`.
    """
    with multiprocessing.Pool() as pool:
        results = pool.starmap(func, zip(iterable, *args))
    return results



# base imports
import io
import re
import os
import getpass
import pandas as pd
import json
import logging
import math
import numpy as np
import gdown
import datetime
import ffmpeg
import shutil
import sqlite3
from tqdm import tqdm
from panoptes_client import Panoptes, panoptes, Subject, SubjectSet
from panoptes_client import Project as zooProject
from ast import literal_eval
from pathlib import Path

# util imports
from kso_utils.project_utils import Project
from kso_utils.db_utils import add_db_info_to_df
import kso_utils.db_utils as db_utils
import kso_utils.movie_utils as movie_utils
import kso_utils.server_utils as server_utils

# Widget imports
import ipywidgets as widgets


# Logging
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)

##########################
# General Zoo purpose functions
##########################


def zoo_credentials():
    zoo_user = getpass.getpass("Enter your Zooniverse user")
    zoo_pass = getpass.getpass("Enter your Zooniverse password")
    return zoo_user, zoo_pass


class AuthenticationError(Exception):
    pass


def connect_zoo_project(project: Project, zoo_cred=False):
    """
    It takes a project name as input, and returns a Zooniverse project object
    zoo_cred is an argument that can pass [username, password] to log in into zooniverse.
    This is used in the automatic tests in gitlab called autotests.py.
    when it is set to False, then the credentials are retrieved from the interacitve widget.
    :param project: the KSO project you are working
    :return: A Zooniverse project object.
    """
    if zoo_cred is False:
        # Save your Zooniverse user name and password.
        zoo_user, zoo_pass = zoo_credentials()
    else:
        zoo_user = zoo_cred[0]
        zoo_pass = zoo_cred[1]

    # Get the project-specific zooniverse number
    project_n = project.Zooniverse_number

    # Connect to the Zooniverse project
    zoo_project = auth_session(zoo_user, zoo_pass, project_n)

    logging.info("Connected to Zooniverse")

    return zoo_project


# Function to authenticate to Zooniverse
def auth_session(username: str, password: str, project_n: int):
    """
    It connects to the Zooniverse with your username and password, and then returns the project number
    you specify

    :param username: Your Zooniverse username
    :param password: your Zooniverse password
    :param project_n: The project number of the project you want to download data from
    :return: The project number of the koster lab
    """

    # Connect to Zooniverse with your username and password
    auth = Panoptes.connect(username=username, password=password)

    if not auth.logged_in:
        raise AuthenticationError("Your credentials are invalid. Please try again.")

    # Specify the project number of the koster lab
    try:
        project = zooProject(int(float(project_n)))
        return project
    except Exception as e:
        logging.error(e)


# Function to retrieve information from Zooniverse
def retrieve_zoo_info(
    project: Project,
    zoo_project,
    zoo_info: str,
    generate_export: bool = False,
):
    """
    This function retrieves the information of interest from Zooniverse and saves it as a pandas data
    frame

    :param project: the kso project object
    :param zoo_project: the Zooniverse project object
    :param zoo_info: a list of the info you want to retrieve from Zooniverse
    :type zoo_info: str
    :param generate_export: boolean determining whether to generate a new export and wait for it to be ready or to just download the latest export
    :return: A dictionary of dataframes.
    """

    # Create an empty dictionary to host the dfs of interest
    info_df = {}

    for info_n in zoo_info:
        logging.info(f"Retrieving {info_n} from Zooniverse")

        try:
            # Get the information of interest from Zooniverse
            if generate_export:
                try:
                    export = zoo_project.get_export(
                        info_n, generate=generate_export, wait=True, wait_timeout=1800
                    )
                except panoptes.PanoptesAPIException:
                    logging.error(
                        "Export generation time out, retrieving the last available information..."
                    )
                    export = zoo_project.get_export(info_n, generate=False)
            else:
                export = zoo_project.get_export(info_n, generate=generate_export)

            # Save the info as pandas data frame
            try:
                export_df = pd.read_csv(
                    io.StringIO(export.content.decode("utf-8")), low_memory=False
                )
            except pd.errors.ParserError as e:
                # Handle the case where a ParserError occurs
                logging.error(f"ParserError occurred while reading CSV data: %s {e}")
                # Optionally, provide a more specific error message or handle the error accordingly
                export_df = None

        except AttributeError:
            logging.info(
                "No connection with Zooniverse, retrieving template info from google drive."
            )
            if info_n == "classifications":
                url = "https://drive.google.com/file/d/1DvJ2nOrG32MR2D7faAJZXMNbEm_ra3rb/view?usp=sharing"
            if info_n == "subjects":
                url = "https://drive.google.com/file/d/18AWRPx3erL25IHekncgKfI_kXHFYAl8e/view?usp=sharing"
            if info_n == "workflows":
                url = "https://drive.google.com/file/d/1bZ6CSxJLxeoX8xVgMU7ZqL76RZDv-09A/view?usp=sharing"
            export = gdown.download(url, info_n + ".csv", quiet=False, fuzzy=True)
            export_df = pd.read_csv(export)

        if len(export_df) > 0:
            # Clear duplicated subjects
            if info_n == "subjects":
                # Select info from each unique subject to avoid loading subject info
                # multiple times (because the subject has been classified in multiple workflows)
                export_df = export_df.drop_duplicates(subset="subject_id", keep="first")

        else:
            raise ValueError(
                "The export is empty. This may be due to a "
                "request time out, please try again in 1 minute."
            )

        # Ensure subject_ids and workflow_ids match db format
        if info_n == "classifications":
            export_df["subject_ids"] = export_df["subject_ids"].astype(np.int64)

        # Add df to dictionary
        info_df[info_n] = export_df
        project.info_df = info_df
        logging.info(f"{info_n} retrieved successfully")

    return info_df


# Function to extract metadata from subjects
def extract_metadata(subj_df: pd.DataFrame):
    """
    The function extracts metadata from a pandas DataFrame and returns two separate DataFrames, one with
    the metadata flattened and one without the metadata.

    :param subj_df: A pandas DataFrame containing subject data, including metadata information in JSON
    format
    :type subj_df: pd.DataFrame
    :return: The function `extract_metadata` returns two dataframes: `subj_df` and `meta_df`. `subj_df`
    is the original input dataframe with the "metadata" and "index" columns dropped, and with the index
    reset. `meta_df` is a new dataframe that contains the flattened metadata information extracted from
    the "metadata" column of the input dataframe.
    """

    # Reset index of df
    subj_df = subj_df.reset_index(drop=True).reset_index()

    # Flatten the metadata information
    meta_df = pd.json_normalize(subj_df.metadata.apply(json.loads))

    # Drop metadata and index columns from original df
    subj_df = subj_df.drop(
        columns=[
            "metadata",
            "index",
        ]
    )

    return subj_df, meta_df


# Function to clean label (no non-alpha characters)
def clean_label(label_string: str):
    label_string = label_string.replace("_", "")
    label_string = label_string.upper()
    label_string = label_string.replace(" ", "")
    pattern = r"[^A-Za-z0-9]+"
    cleaned_string = re.sub(pattern, "", label_string)
    return cleaned_string


##########################
# Workflow-specific functions
##########################


class WidgetWorkflowSelection(widgets.VBox):
    def __init__(self, workflows_df: pd.DataFrame):
        """
        The function creates a widget that allows the user to select which workflows to run

        :param workflows_df: the dataframe of workflows
        """

        # Estimate the maximum number of workflows available
        max_workflows = workflows_df.display_name.nunique()

        self.workflows_df = workflows_df
        self.widget_count = widgets.BoundedIntText(
            value=0,
            min=0,
            max=max_workflows,
            description="Number of workflows:",
            display="flex",
            flex_flow="column",
            align_items="stretch",
            style={"description_width": "initial"},
        )
        self.bool_widget_holder = widgets.HBox(
            layout=widgets.Layout(
                width="70%", display="inline-flex", flex_flow="row wrap"
            )
        )
        children = [
            self.widget_count,
            self.bool_widget_holder,
        ]
        self.widget_count.observe(self._add_bool_widgets, names=["value"])
        super().__init__(children=children)

    def _add_bool_widgets(self, widg):
        num_bools = widg["new"]
        new_widgets = []
        for _ in range(num_bools):
            from kso_utils.widgets import choose_workflows

            new_widget = choose_workflows(self.workflows_df)
            for wdgt in new_widget:
                wdgt.description = wdgt.description + f" #{_}"
            new_widgets.extend(new_widget)
        self.bool_widget_holder.children = tuple(new_widgets)

    @property
    def checks(self):
        return {w.description: w.value for w in self.bool_widget_holder.children}


def get_workflow_labels(
    workflow_df: pd.DataFrame, workflow_id: int, workflow_version: int
):
    """
    > This function takes a df of workflows of interest and retrieves the labels and common names of the choices cit scientists have in a survey task in Zooniverse.
    the function is a modified version of the 'get_workflow_info' function by @lcjohnso
    https://github.com/zooniverse/Data-digging/blob/6e9dc5db6f6125316616c4b04ae5fc4223826a25/scripts_GeneralPython/get_workflow_info.pybiological observations classified by citizen scientists, biologists or ML algorithms and returns a df of species occurrences to publish in GBIF/OBIS.
    :param workflow_df: df of the workflows of the Zooniverse project of interest,
    :param workflow_id: integer of the workflow id of interest,
    :param workflow_version: integer of the workflow version of interest.
    :return: a df with the common name and label of the annotations for the workflow.
    """
    # initialize the output
    workflow_info = {}

    # parse the tasks column as a json so we can work with it (it just loads as a string)
    workflow_df["tasks_json"] = [json.loads(q) for q in workflow_df["tasks"]]
    workflow_df["strings_json"] = [json.loads(q) for q in workflow_df["strings"]]

    # identify the row of the workflow dataframe we want to extract
    is_theworkflow = (workflow_df["workflow_id"] == workflow_id) & (
        workflow_df["version"] == workflow_version
    )

    # extract it
    theworkflow = workflow_df[is_theworkflow]

    # pandas is a little weird about accessing stuff sometimes
    # we should only have 1 row in theworkflow but the row index will be retained
    # from the full workflow_df, so we need to figure out what it is
    i_wf = theworkflow.index[0]

    # extract the tasks as a json
    tasks = theworkflow["tasks_json"][i_wf]
    strings = theworkflow["strings_json"][i_wf]

    workflow_info = tasks.copy()

    tasknames = workflow_info.keys()
    workflow_info["tasknames"] = tasknames

    # now that we've extracted the actual task names, add the first task
    workflow_info["first_task"] = theworkflow["first_task"].values[0]

    # now join workflow structure to workflow label content for each task
    for task in tasknames:
        # Check if the task has multiple choices
        if isinstance(workflow_info[task], dict):
            # Create an empty dictionary to host the dfs of interest
            label_common_name_dict = {"commonName": [], "label": []}
            for choice in workflow_info[task]["choices"]:
                label_common_name_dict["label"].append(choice)
                choice_name = strings[workflow_info[task]["choices"][choice]["label"]]
                label_common_name_dict["commonName"].append(choice_name)

        if task == "T0":
            break

    return pd.DataFrame.from_dict(label_common_name_dict)


##########################
# Classification-specific functions
##########################


# Flatten the classifications provided the cit. scientists
def process_zoo_classifications(
    project: Project,
    server_connection: dict,
    db_connection,
    workflow_widget_checks,
    workflows_df: pd.DataFrame,
    subjects_df: pd.DataFrame,
    csv_paths: dict,
    classifications_data: pd.DataFrame,
    subject_type: str,
):
    """
    This function select subjects from the workflows of interest, populates the sql subjects table, process the classifications received and returns a dataframe of annotations.

    :param project: the project object
    :param server_connection: A dictionary with the client and sftp_client
    :param db_connection: SQL connection object
    :param workflow_widget_checks: the widget with information of the selected workflows
    :param workflows_df: dataframe with the project workflows information (retrieved from Zooniverse)
    :param subjects_df: dataframe with the project subjects information (retrieved from Zooniverse)
    :param csv_paths: a dictionary with the paths of the csvs used to initiate the db
    :param classifications_data: the dataframe of classifications from the Zooniverse API
    :param subject_type: This is the type of subject you want to retrieve classifications for. This
           can be either "clip" or "frame"
    """

    # Store the names of the workflows and their versions
    workflow_names, workflow_versions = [], []
    for i in range(0, len(workflow_widget_checks), 3):
        workflow_names.append(list(workflow_widget_checks.values())[i])
        workflow_versions.append(list(workflow_widget_checks.values())[i + 2])

    # Get the ids of the workflows of interest
    selected_zoo_workflows = [
        workflows_df[workflows_df.display_name == wf_name].workflow_id.unique()[0]
        for wf_name in workflow_names
    ]

    if not isinstance(selected_zoo_workflows, list):
        selected_zoo_workflows = literal_eval(selected_zoo_workflows)

    # Classifications selected##############
    # Select only classification from the workflows of interest
    classes_df = classifications_data[
        classifications_data["workflow_id"].isin(selected_zoo_workflows)
    ].copy()

    # Get the unique subject IDs from the classes DataFrame
    unique_subject_ids = classes_df["subject_ids"].unique()

    # Count the number of unique subject IDs
    number_of_unique_subjects = len(unique_subject_ids)

    # Report the number of classifications retrieved
    logging.info(
        f"{classes_df.shape[0]:,} Zooniverse classifications have been retrieved"
        f" from {number_of_unique_subjects:,} subjects"
    )

    # Subjects selected########################
    # Select only subjects from the workflows of interest
    subjects_series = subjects_df[
        subjects_df["subject_id"].isin(unique_subject_ids)
    ].copy()

    from kso_utils.db_utils import drop_table

    # Safely remove subjects table
    drop_table(conn=db_connection, table_name="subjects")

    if len(subjects_series) > 0:
        # Fill or re-fill subjects table
        populate_subjects(project, server_connection, db_connection, subjects_series)
    else:
        logging.error("No subjects to populate database from the workflows selected.")

    # Combine classifications and subject info#####
    # Combine the classifications and subjects dataframes
    classes_df = add_db_info_to_df(
        project,
        db_connection,
        csv_paths,
        classes_df,
        "subjects",
        "id, subject_type, filename, clip_start_time, clip_end_time, frame_exp_sp_id, frame_number, subject_set_id, https_location, movie_id",
    )

    # Flatten the classifications provided the cit. scientists

    # Create an empty list to store the annotations
    rows_list = []

    # Loop through each classification submitted by the users
    for index, row in classes_df.iterrows():
        # Load annotations as json format
        annotations = json.loads(row["annotations"])

        # Process clip annotations
        if subject_type == "clip":
            # Select the information from the species identification task
            if project.Zooniverse_number == 9747:
                from kso_utils.koster_utils import process_clips_koster

                rows_list = process_clips_koster(
                    annotations, row["classification_id"], rows_list
                )

            # Check if the Zooniverse project is the Spyfish
            if project.Project_name == "Spyfish_Aotearoa":
                from kso_utils.spyfish_utils import process_clips_spyfish

                rows_list = process_clips_spyfish(
                    annotations, row["classification_id"], rows_list
                )

            # Process clips as the default method
            else:
                rows_list = process_clips_default(
                    annotations, row["classification_id"], rows_list
                )

        # Process frame classifications
        elif subject_type == "frame":
            # Select the information from all the labelled animals (e.g. task = T0)
            for ann_i in annotations:
                # Temporary fix (find out what T2 is and if it is always relevant)
                if ann_i["task"] in ["T0", "T2"]:
                    if ann_i["value"] == []:
                        # Specify the frame was classified as empty
                        choice_i = {
                            "classification_id": row["classification_id"],
                            "x": None,
                            "y": None,
                            "w": None,
                            "h": None,
                            "label": "empty",
                        }
                        rows_list.append(choice_i)

                    else:
                        # Select each species annotated and flatten the relevant answers
                        for i in ann_i["value"]:
                            choice_i = {
                                "classification_id": row["classification_id"],
                                "x": (
                                    int(i["x"])
                                    if "x" in i
                                    else (
                                        int(i["x_center"]) if "x_center" in i else None
                                    )
                                ),
                                "y": (
                                    int(i["y"])
                                    if "y" in i
                                    else (
                                        int(i["y_center"]) if "y_center" in i else None
                                    )
                                ),
                                "w": int(i["width"]) if "width" in i else None,
                                "h": int(i["height"]) if "height" in i else None,
                                "poly_points": i["points"] if "points" in i else None,
                                "label": (
                                    str(i["tool_label"])
                                    if "tool_label" in i and i["tool_label"] is not None
                                    else (
                                        str(i["toolIndex"])
                                        if "toolIndex" in i
                                        else None
                                    )
                                ),
                            }
                            rows_list.append(choice_i)

        else:
            logging.error("The subject_type is not valid")

    # Specify the cols specific to each subject type
    if subject_type == "clip":
        subject_cols = ["first_seen", "how_many"]
    else:
        subject_cols = ["x", "y", "w", "h", "poly_points"]

    # Combine common columns for the flattened annotations with subject type specific
    annot_cols = ["classification_id", "label"]
    annot_cols.extend(subject_cols)

    # Create a data frame of the flatten annotations
    flat_annot_df = pd.DataFrame(rows_list, columns=annot_cols)

    # Ensure the empty cells are replace with NAN
    flat_annot_df[subject_cols] = (
        flat_annot_df[subject_cols]
        .astype(str)
        .apply(lambda x: x.str.strip())
        .replace("", np.nan)
        .replace("None", np.nan)
    )

    # Ensure the subject type specific columns are numeric
    if subject_type == "frame":
        coord_cols = ["x", "y", "w", "h"]
        flat_annot_df[coord_cols] = flat_annot_df[coord_cols].astype("float64")

    # Combine the flatten the classifications with the subject information

    # Add subject information to each annotation
    annot_df = pd.merge(
        flat_annot_df,
        classes_df.drop(columns=["annotations"]),
        how="left",
        on="classification_id",
    )

    # Specify relevant columns
    annot_cols.extend(
        [
            "https_location",
            "subject_type",
            "subject_ids",
            "workflow_id",
            "workflow_name",
            "workflow_version",
            "user_name",
            "filename",
            "clip_start_time",
            "clip_end_time",
            "frame_exp_sp_id",
            "frame_number",
            "movie_id",
        ]
    )

    # Select only relevant columns
    annot_df = annot_df[annot_cols]

    # Select only classification from the workflows of interest
    annot_df = annot_df[annot_df.workflow_id.isin(selected_zoo_workflows)].copy()

    # Report the number of annotations flattened
    logging.info(
        f"{annot_df.shape[0]:,} Zooniverse annotations have been flattened"
        f" from {annot_df.subject_ids.nunique():,} subjects"
    )

    return pd.DataFrame(annot_df)


def process_clips_default(annotations: pd.DataFrame, row_class_id, rows_list: list):
    """
    For each annotation, if the task is T0, then for each species annotated, flatten the relevant
    answers and save the species of choice, class and subject id.

    :param annotations: the list of annotations for a given subject
    :param row_class_id: the classification id
    :param rows_list: a list of dictionaries, each dictionary is a row in the output dataframe
    :return: A list of dictionaries, each dictionary containing the classification id, the label, the first seen time and the number of individuals.

    """

    for ann_i in annotations:
        if ann_i["task"] == "T0":
            # Select each species annotated and flatten the relevant answers
            for value_i in ann_i["value"]:
                choice_i = {}
                # If choice = 'nothing here', set follow-up answers to blank
                if value_i["choice"] == "NOTHINGHERE":
                    f_time = ""
                    inds = ""
                # If choice = species, flatten follow-up answers
                else:
                    answers = value_i["answers"]
                    for k in answers.keys():
                        if "EARLIESTPOINT" in k:
                            f_time = answers[k].replace("S", "")
                        if "HOWMANY" in k:
                            inds = answers[k]
                            # Deal with +20 fish options
                            if inds == "2030":
                                inds = "25"
                            if inds == "3040":
                                inds = "35"
                        elif "EARLIESTPOINT" not in k and "HOWMANY" not in k:
                            f_time, inds = None, None

                # Save the species of choice, class and subject id
                choice_i.update(
                    {
                        "classification_id": row_class_id,
                        "label": value_i["choice"],
                        "first_seen": f_time,
                        "how_many": inds,
                    }
                )

                rows_list.append(choice_i)

    return rows_list


def aggregate_classifications(
    project: Project,
    processed_classifications: pd.DataFrame,
    subject_type: str,
    agg_params: list,
):
    """
    We take the processed citizen scientits classifications and
    aggregated them based on aggregration paramaeters

    :param df: the processed classifications dataframe
    :param subject_type: the type of subject, either "frame" or "clip"
    :param agg_params: list of parameters for the aggregation
    :return: the aggregated classifications classifications.
    """

    # Check if we have the right number of parameters for the type of subject
    if subject_type == "clip":
        subject_parameters = 2
    elif subject_type == "frame":
        subject_parameters = 5

    if not len(agg_params) == subject_parameters:
        logging.error("Incorrect agg_params length for subject type")
        return

    logging.info("Aggregating the classifications")

    # We take the processed_classifications and aggregate them.
    if subject_type == "frame":
        # Get the aggregation parameters
        agg_users, min_users, agg_obj, agg_iou, agg_iua = [i.value for i in agg_params]
        # Report selected parameters
        logging.info(
            f"Aggregation parameters are: Agg. threshold "
            f"{agg_users} "
            f"Min. users "
            f"{min_users} "
            f"Obj threshold "
            f"{agg_obj} "
            f"IOU "
            f"{agg_iou} "
            f"Int. agg. "
            f"{agg_iua} "
        )

        # Aggregate frames based on volunteer consensus of the labels
        agg_labels_df = aggregate_labels(
            processed_classifications, agg_users, min_users
        )

        # Get rid of the "empty" labels if other species are among the volunteer consensus
        agg_labels_df = agg_labels_df[
            ~((agg_labels_df["class_n_agg"] > 1) & (agg_labels_df["label"] == "empty"))
        ]

        # Select frames aggregated only as empty (to add them later)
        empty_class_df = agg_labels_df[agg_labels_df["label"] == "empty"]

        # Temporary exclude frames already aggregated as empty
        agg_labels_df = agg_labels_df[
            ~agg_labels_df["classification_id"].isin(
                empty_class_df.classification_id.unique()
            )
        ]

        # Map the position of the annotation parameters
        col_list = list(agg_labels_df.columns)
        x_pos, y_pos, w_pos, h_pos, class_id_pos, subject_id_pos = (
            col_list.index("x"),
            col_list.index("y"),
            col_list.index("w"),
            col_list.index("h"),
            col_list.index("classification_id"),
            col_list.index("subject_ids"),
        )

        # Get prepared annotations
        new_rows = []

        # Specify the cols to group the annotations by
        group_cols = ["subject_ids", "label"]

        # loop through each group and estimate the overlapped annotations
        for name, group in agg_labels_df.groupby(group_cols):
            subj_id, label = name
            total_users = agg_labels_df[
                (agg_labels_df.subject_ids == subj_id) & (agg_labels_df.label == label)
            ]["user_name"].nunique()

            # Filter bboxes using IOU metric (essentially a consensus metric)
            # Keep only bboxes where mean overlap exceeds this threshold
            from kso_utils.frame_utils import filter_bboxes

            indices, new_group = filter_bboxes(
                total_users=total_users,
                users=[i[class_id_pos] for i in group.values],
                bboxes=[
                    np.array([i[x_pos], i[y_pos], i[w_pos], i[h_pos]])
                    for i in group.values
                ],
                obj=agg_obj,
                eps=agg_iou,
                iua=agg_iua,
            )

            subject_ids = [i[subject_id_pos] for i in group.values[indices]]

            for ix, box in zip(subject_ids, new_group):
                new_rows.append(
                    (
                        label,
                        ix,
                    )
                    + tuple(box)
                )

        # Specify the names of the relevant columns for the agggregated df
        cols_agg_class_df = [
            "label",
            "subject_ids",
            "x",
            "y",
            "w",
            "h",
        ]

        # Combine the new rows into a df
        agg_class_df = pd.DataFrame(
            new_rows,
            columns=cols_agg_class_df,
        )

        # Clear some weird labels
        agg_class_df["label"] = agg_class_df["label"].apply(
            lambda x: x.split("(")[0].strip()
        )

        # Select only relevant columns from the empty df
        empty_class_df = empty_class_df[cols_agg_class_df]

        # Add the frames aggregated as "empty"
        agg_class_df = pd.concat([agg_class_df, empty_class_df])

        # Specify the columns from the processed_classifications to drop
        # to avoid duplicates
        cols_processed_classifications = [
            x for x in cols_agg_class_df if x != "subject_ids"
        ]

        # temporarily drop the unnecessary cols from the processed classifications
        temp_processed_classifications = processed_classifications.drop(
            columns=cols_processed_classifications
        )

        # Add the subject,site,movie.. info related to the subject id
        # from the processed classifications
        agg_class_df = agg_class_df.merge(
            temp_processed_classifications, on="subject_ids"
        )

    else:
        # Get the aggregation parameters
        if not isinstance(agg_params, list):
            agg_users, min_users = [i.value for i in agg_params]
        else:
            agg_users, min_users = agg_params

        # Aggregate clips based on their labels
        agg_class_df = aggregate_labels(processed_classifications, agg_users, min_users)

    # Drop unnecessary columns
    agg_class_df = agg_class_df.drop(columns=["classification_id", "user_name"])

    # Keep only one raw per subject and label
    agg_class_df = agg_class_df.drop_duplicates()

    logging.info(
        f"{agg_class_df.shape[0]}"
        " classifications aggregated out of "
        f"{processed_classifications.subject_ids.nunique()}"
        " unique subjects available"
    )

    return agg_class_df


def aggregate_labels(
    processed_classifications: pd.DataFrame, agg_users: float, min_users: int
):
    """
    > This function takes a dataframe of classifications and returns a dataframe of classifications that
    have been filtered by the number of users that classified each subject and the proportion of users
    that agreed on their annotations

    :param processed_classifications: the dataframe of all the classifications
    :param agg_users: the proportion of users that must agree on a classification for it to be included
           in the final dataset
    :param min_users: The minimum number of users that must have classified a subject for it to be
           included in the final dataset
    :return: a dataframe with the aggregated labels.
    """
    # Calculate the number of users that classified each subject
    processed_classifications["n_users"] = processed_classifications.groupby(
        "subject_ids"
    )["classification_id"].transform("nunique")

    # Select classifications with at least n different user classifications
    processed_classifications = processed_classifications[
        processed_classifications.n_users >= min_users
    ].reset_index(drop=True)

    # Calculate the proportion of unique classifications (it can have multiple annotations) per subject
    processed_classifications["class_n"] = processed_classifications.groupby(
        ["subject_ids", "label"]
    )["classification_id"].transform("nunique")

    # Calculate the proportion of users that agreed on their annotations
    processed_classifications["aggreg"] = (
        processed_classifications.class_n / processed_classifications.n_users
    )

    # Select annotations based on agreement threshold
    agg_class_df = processed_classifications[
        processed_classifications.aggreg >= agg_users
    ].reset_index(drop=True)

    # Specify the columns to group by and calculate the median to
    # of the second where the animal/object is and number of animals
    list_cols_group = ["subject_ids", "label"]
    cols_median = ["how_many", "first_seen"]

    # If the df has columns to calculate the median (usually clips)
    if all(item in agg_class_df.columns for item in cols_median):
        # Prevent issues when trying to calculate the median of empty cells
        agg_class_df[cols_median] = agg_class_df[cols_median].fillna(0)

        # Extract the median of the second where the animal/object is and number of animals
        agg_class_df[cols_median] = (
            agg_class_df.groupby(list_cols_group, observed=True)[cols_median]
            .transform("median")
            .astype(int)
        )

    # Calculate the number of unique classifications aggregated per subject
    agg_class_df["class_n_agg"] = agg_class_df.groupby(["subject_ids"])[
        "label"
    ].transform("nunique")

    return agg_class_df


def add_subject_site_movie_info_to_class(
    project: Project,
    db_connection,
    csv_paths: dict,
    class_df: pd.DataFrame,
):
    """
    It takes a dataframe of clips or frames, and adds metadata about the site and project to it
    :param df: the dataframe with the media to upload
    :param project: the project object
    :param species_list: a list of the species that should be on the frames
    :param csv_paths: a dictionary with the paths of the csvs used to initiate the db
    :param class_df: a data frame with the classifications
    :param aggregated: a boolean value specifyin if the classifications are aggregated or not
    """

    # Combine the aggregated clips and movies dataframes
    class_df = add_db_info_to_df(
        project,
        db_connection,
        csv_paths,
        class_df,
        "movies",
        "id, created_on, fps, duration, sampling_start, sampling_end, author, site_id, fpath",
    )

    # Combine the aggregated clips and sites dataframes
    class_df = add_db_info_to_df(project, db_connection, csv_paths, class_df, "sites")

    return class_df


def launch_classifications_table(agg_class_df: pd.DataFrame, subject_type: str):
    """
    It takes in a dataframe of aggregated classifications and a subject type, and returns a dataframe
    with the columns "subject_ids", "label", "how_many", and "first_seen"

    :param agg_class_df: the dataframe that you want to launch
    :param subject_type: "clip" or "subject"
    """
    if subject_type == "clip":
        a = agg_class_df[["subject_ids", "label", "how_many", "first_seen"]]
    else:
        a = agg_class_df

    return a


##########################
# Subject-specific functions
##########################


def get_workflow_ids(workflows_df: pd.DataFrame, workflow_names: list):
    # The function that takes a list of workflow names and returns a list of workflow
    # ids.
    return [
        workflows_df[workflows_df.display_name == wf_name].workflow_id.unique()[0]
        for wf_name in workflow_names
    ]


def get_classifications(
    project: Project,
    conn: sqlite3.Connection,
    workflow_dict: dict,
    workflows_df: pd.DataFrame,
    subj_type: str,
    class_df: pd.DataFrame,
):
    """
    It takes in a dictionary of workflows, a dataframe of workflows, the type of subject (frame or
    clip), a dataframe of classifications, the path to the database, and the project name. It returns a
    dataframe of classifications

    :param project: the project object
    :param conn: SQL connection object
    :param workflow_dict: a dictionary of the workflows you want to retrieve classifications for. The
        keys are the workflow names, and the values are the workflow IDs, workflow versions, and the minimum
        number of classifications per subject
    :type workflow_dict: dict
    :param workflows_df: the dataframe of workflows from the Zooniverse project
    :type workflows_df: pd.DataFrame
    :param subj_type: "frame" or "clip"
    :param class_df: the dataframe of classifications from the database
    :return: A dataframe with the classifications for the specified project and workflow.
    """

    names, workflow_versions = [], []
    for i in range(0, len(workflow_dict), 3):
        names.append(list(workflow_dict.values())[i])
        workflow_versions.append(list(workflow_dict.values())[i + 2])

    workflow_ids = get_workflow_ids(workflows_df, names)

    # Filter classifications of interest
    classes = []
    for id, version in zip(workflow_ids, workflow_versions):
        class_df_id = class_df[
            (class_df.workflow_id == id) & (class_df.workflow_version >= version)
        ].reset_index(drop=True)
        classes.append(class_df_id)
    classes_df = pd.concat(classes)

    # Add information about the subject
    # Query id and subject type from the subjects table
    subjects_df = db_utils.get_df_from_db_table(conn, "subjects")

    if subj_type == "frame":
        # Select only frame subjects
        subjects_df = subjects_df[subjects_df["subject_type"] == "frame"]

        # Select columns relevant for frame subjects
        subjects_df = subjects_df[
            [
                "id",
                "subject_type",
                "https_location",
                "filename",
                "frame_number",
                "movie_id",
            ]
        ]

    else:
        # Select only clip subjects
        subjects_df = subjects_df[subjects_df["subject_type"] == "clip"]

        # Select columns relevant for clip subjects
        subjects_df = subjects_df[
            [
                "id",
                "subject_type",
                "https_location",
                "filename",
                "clip_start_time",
                "movie_id",
            ]
        ]

    # Ensure id format matches classification's subject_id
    classes_df["subject_ids"] = classes_df["subject_ids"].astype("Int64")
    subjects_df["id"] = subjects_df["id"].astype("Int64")

    # Add subject information based on subject_ids
    classes_df = pd.merge(
        classes_df,
        subjects_df,
        how="left",
        left_on="subject_ids",
        right_on="id",
    )

    if classes_df[["subject_type", "https_location"]].isna().any().any():
        # Exclude classifications from missing subjects
        filtered_class_df = classes_df.dropna(
            subset=["subject_type", "https_location"], how="any"
        ).reset_index(drop=True)

        # Report on the issue
        logging.info(
            f"There are {(classes_df.shape[0]-filtered_class_df.shape[0])}"
            f" classifications out of {classes_df.shape[0]}"
            f" missing subject info. Maybe the subjects have been removed from Zooniverse?"
        )

        classes_df = filtered_class_df

    logging.info(
        f"{classes_df.shape[0]} Zooniverse classifications have been retrieved"
    )

    return classes_df


def populate_subjects(
    project: Project,
    server_connection: dict,
    db_connection,
    subjects: pd.DataFrame,
):
    """
    Populate the subjects table with the subject metadata
    :param project: the project object
    :param server_connection: A dictionary with the client and sftp_client
    :param subjects: a dataframe of the subset of subjects from the workflows of interest
    :param db_connection: SQL connection object

    """

    # Check if the Zooniverse project is the KSO
    if project.Project_name == "Koster_Seafloor_Obs":
        from kso_utils.koster_utils import process_koster_subjects

        subjects = process_koster_subjects(subjects, db_connection)

    else:
        # Extract metadata from uploaded subjects
        subjects_df, subjects_meta = extract_metadata(subjects)

        # Combine metadata info with the subjects df
        subjects = pd.concat([subjects_df, subjects_meta], axis=1)

        # Check if the Zooniverse project is the Spyfish
        if project.Project_name == "Spyfish_Aotearoa":
            from kso_utils.spyfish_utils import process_spyfish_subjects

            # Process the spyfish movies
            subjects = process_spyfish_subjects(
                project, server_connection, subjects, db_connection
            )

    # Ensure that subject_ids are not duplicated by workflow
    subjects = subjects.drop_duplicates(subset="subject_id")

    # Rename columns that start with "#"
    hash_columns = {col: col.replace("#", "") for col in subjects.columns}

    # Rename columns to match the db format
    subjects = subjects.rename(columns=hash_columns)

    # Fix subject_type missing info and non-standard format
    # Fix weird bug where Subject_type is used instead of subject_type for the column name for some clips
    if "Subject_type" in subjects.columns and "subject_type" in subjects.columns:
        subjects["subject_type"] = subjects[["subject_type", "Subject_type"]].apply(
            lambda x: x[1] if isinstance(x[1], str) else x[0], 1
        )
        subjects.drop(columns=["Subject_type"], inplace=True)

    # Fix similar weird bug to above (Jannes)
    if "upl_seconds" in subjects.columns and "clip_start_time" in subjects.columns:
        subjects["clip_start_time"] = subjects[
            ["clip_start_time", "upl_seconds"]
        ].apply(lambda x: x[1] if not np.isnan(x[1]) else x[0], 1)
        subjects.drop(columns=["upl_seconds"], inplace=True)

    # Fix similar weird bug to above (Jannes)
    if "VideoFilename" in subjects.columns and "filename" in subjects.columns:
        subjects["filename"] = subjects[["filename", "VideoFilename"]].apply(
            lambda x: x[1] if isinstance(x[1], str) else x[0], 1
        )
        subjects.drop(columns=["VideoFilename"], inplace=True)

    # Rename common non-standard column names
    rename_cols = {
        "subject_id": "id",
        "VideoFilename": "filename",
        "Filename": "filename",
        "upl_seconds": "clip_start_time",
        "Subject_type": "subject_type",
    }

    # Rename columns to match the db format
    subjects = subjects.rename(columns=rename_cols)

    if "subject_type" in subjects.columns:
        # Avoid having two subject_type columns (one from Zoo one from the db)
        if isinstance(subjects["subject_type"], pd.DataFrame):
            subjects["subject_type0"] = subjects["subject_type"].iloc[:, 0]
            subjects["subject_type1"] = subjects["subject_type"].iloc[:, 1]
        else:
            subjects["subject_type0"] = subjects["subject_type"]
            subjects["subject_type1"] = subjects["subject_type"]

        # Update with non-empty values
        subjects["combined_subject_type"] = subjects["subject_type0"].combine_first(
            subjects["subject_type1"]
        )

        # Drop unnecesary cols
        subjects.drop(
            columns=["subject_type", "subject_type0", "subject_type1"], inplace=True
        )

        # Rename the combined subject type
        subjects = subjects.rename(columns={"combined_subject_type": "subject_type"})
    else:
        subjects["subject_type"] = "frame"

    # Extract the html location of the subjects
    subjects["https_location"] = subjects["locations"].apply(
        lambda x: literal_eval(x)["0"]
    )

    # Check for empty values in subject_type
    if subjects["subject_type"].isna().any():
        # Manually write the subject type based on the extension of the subject location (link in Zooniverse)
        from kso_utils.movie_utils import get_movie_extensions

        # Calculate the number of subjects with empty subject type info
        num_updated_empty = subjects[subjects["subject_type"].isna()].shape[0]

        # Update empty values with "clip" based on https_location extension
        subjects.loc[
            subjects["subject_type"].isna(),
            "subject_type",
        ] = subjects[
            subjects["subject_type"].isna()
        ]["https_location"].apply(
            lambda url: (
                "clip"
                if any(url.endswith(ext) for ext in get_movie_extensions())
                else "frame"
            )
        )
        logging.info(
            f"Updated {num_updated_empty} empty rows with subject_type based on extension."
        )

    # Ensure only subjects with the right format get populated
    right_types = ["frame", "clip"]
    if subjects[subjects.subject_type.isin(right_types)].all().all():
        # Count the number of rows to be excluded
        logging.info(
            f"{subjects[~subjects.subject_type.isin(right_types)].shape[0]}"
            f" subjects missing the right subject_type info"
        )

        # Select only rows with the right subject_type info
        subjects = subjects[subjects["subject_type"].isin(right_types)]

    if "movie_id" not in subjects.columns:
        from kso_utils.db_utils import get_df_from_db_table

        # Match site code to name from movies sql and get movie_id to save it as "movie_id"
        # Query id and filenames from the movies table
        movies_df = get_df_from_db_table(db_connection, "movies")[["id", "filename"]]

        # Rename columns to match subject df
        movies_df = movies_df.rename(columns={"id": "movie_id"})

        # Reference the movienames with the id movies table
        subjects = pd.merge(subjects, movies_df, how="left", on="filename")

    # Warn users if original movies can't be traced
    # back using the subject information
    if subjects.movie_id.isnull().values.any():
        # Store the classifications that don't have movie info
        movie_missing = subjects[subjects.movie_id.isnull()]

        # Report the number of classifications retrieved
        logging.info(
            f"The original movies of {movie_missing.id.nunique():,}"
            f" subjects couldn't be traced back, ensure the subject information"
            f" related to the orignal movie matches the information in the movies.csv,"
            #             f" The trouble subjects are: {movie_missing.id.unique()}"
            f" The trouble filename of the movies are: {movie_missing.filename.unique()}"
        )

    if subjects["subject_type"].value_counts().idxmax() == "clip":
        # Calculate the clip_end_time
        subjects["clip_end_time"] = (
            subjects["clip_start_time"] + subjects["clip_length"]
        )

    if subjects["subject_type"].value_counts().idxmax() == "frame":
        # Ensure only one value per expected species id
        # this value is not crucial as we use the labels
        # to process the actual classifications
        # Modify the DataFrame to retain only the first value from each list
        if "frame_exp_sp_id" in subjects.columns:
            subjects["frame_exp_sp_id"] = subjects["frame_exp_sp_id"].apply(
                lambda x: x[0] if isinstance(x, list) and len(x) > 0 else x
            )

    # Get the names required for the subject table
    from kso_utils.db_utils import get_column_names_db

    required_cols = list(get_column_names_db(db_connection, "subjects").values())

    # Convert lists to sets and find the difference
    difference = list(set(required_cols) - set(subjects.columns))

    if difference:
        logging.info(f"Creating empty values for the {difference} columns")
        # Create columns to match schema if they don't exist
        [subjects.__setitem__(i, subjects.get(i, np.nan)) for i in difference]

    # Set the subject columns in the right order
    subjects = subjects[required_cols]

    from kso_utils.db_utils import test_table, add_to_table, get_df_from_db_table

    # Test table validity
    test_table(subjects, "subjects", keys=["id"])

    # Add values to subjects
    add_to_table(
        conn=db_connection,
        table_name="subjects",
        values=[tuple(i) for i in subjects.values],
        num_fields=11,
    )

    # log how many subjects are in the db
    # Query id and subject type from the subjects table
    subjects_df = get_df_from_db_table(db_connection, "subjects")

    logging.debug(
        f"The subjects database has now a total of " f"{subjects_df.shape[0]} subjects"
    )


##########################
# Upload clips functions
##########################


def check_movies_uploaded_zoo(project: Project, db_connection, selected_movies: list):
    """
    This function takes in a movie name and a dictionary containing the path to the database and returns
    a boolean value indicating whether the movie has already been uploaded to Zooniverse

    :param project: the project object
    :param selected_movies: the name of the movie(s) you want to check
    :type selected_movies: list
    :param db_connection: SQL connection object
    """
    from kso_utils.db_utils import get_df_from_db_table

    # Query info about the clip subjects uploaded to Zooniverse from the db
    subjects_df = get_df_from_db_table(conn=db_connection, table_name="subjects")

    # Select only columns of interest
    subjects_df = subjects_df[
        [
            "id",
            "subject_type",
            "filename",
            "clip_start_time",
            "clip_end_time",
            "movie_id",
        ]
    ]

    # Select only clip subjects
    subjects_df = subjects_df[subjects_df["subject_type"] == "clip"]

    # Save the video filenames of the clips uploaded to Zooniverse
    clips_uploaded = subjects_df[
        subjects_df["filename"].apply(
            lambda x: any(movie in x for movie in selected_movies)
        )
    ]

    if clips_uploaded.empty:
        logging.info(f"{selected_movies} has not been uploaded to Zooniverse yet")
    else:
        logging.info(f"{clips_uploaded} clips have already been uploaded.")


# Function to extract the videos
def extract_clips(
    movie_path: str,
    clip_length: int,
    upl_second_i: int,
    output_clip_path: str,
    modification_details: dict,
    gpu_available: bool,
    remove_audio: bool = True,
):
    """
    This function takes in a movie path, a clip length, a starting second index, an output clip path, a
    dictionary of modification details, and a boolean indicating whether a GPU is available. It then
    extracts a clip from the movie, and applies the modifications specified in the dictionary.

    The function is written in such a way that it can be used to extract clips from a movie, and apply
    modifications to the clips.

    :param movie_path: The path to the movie file
    :param clip_length: The length of the clip in seconds
    :param upl_second_i: The second in the video to start the clip
    :param output_clip_path: The path to the output clip
    :param modification_details: a dictionary of dictionaries, where each dictionary contains the
           details of a modification to be made to the video. The keys of the dictionary are the names of the
           modifications, and the values are dictionaries containing the details of the modification.
    :param gpu_available: If you have a GPU, set this to True. If you don't, set it to False
    :param remove_audio: Boolen parameter specifying if remove the audio or not
    """
    # Ensure the clip doesn't already exist
    output_clip_path_check = Path(output_clip_path)

    if output_clip_path_check.exists():
        logging.info(
            f"The clip {output_clip_path} already exists, please remove it and run the code again."
        )
        pass

    from kso_utils.movie_utils import check_ffmpeg_availability

    # Ensure ffmpeg is installed
    if not check_ffmpeg_availability():
        raise FileNotFoundError(
            "ffmpeg is not found in your system's PATH. Please install or add its path to the PATH environment variable."
        )

    # Set up input and output default prompts
    input_path = movie_path
    output_path = str(output_clip_path)

    input_options = {}
    output_options = {}

    # Add GPU-related options if available
    if gpu_available:
        input_options["hwaccel"] = "cuda"
        output_options["c"] = "copy"

    # Add audio-related options if needed
    if remove_audio:
        output_options["an"] = None

    # Access modifications from clip_modification_widget
    if modification_details:
        # Set up modification
        for (
            modif_number,
            modification_details_dict,
        ) in modification_details.checks.items():
            if "filter" in modification_details_dict:
                output_options["vf"] = modification_details_dict["filter"]
            elif "b:v" in modification_details_dict:
                # Specify the select size (MB) of the clip
                output_options["b:v"] = modification_details_dict["b:v"]

                # add the cuda-compatible vcodec
                if gpu_available:
                    # Remove the copy format argument
                    del output_options["c"]
                    output_options["c:v"] = "h264_nvenc"
                    logging.info("Compressing using h264_nvenc")

    input_options["ss"] = upl_second_i
    input_options["t"] = clip_length

    # Run the ffmpeg clip extraction code
    try:
        ffmpeg.input(input_path, **input_options).output(
            output_path, **output_options
        ).run(overwrite_output=True)

    except ffmpeg.Error as e:
        logging.error("ffmpeg error occurred.")
        logging.error(f"stderr: {e}")
        if e.stdout is not None:
            logging.error(f"stdout: {e.stdout.decode('utf8')}")
        if e.stderr is not None:
            logging.error(f"stderr: {e.stderr.decode('utf8')}")
        raise e

    # Ensure the clip was extracted
    if not os.path.exists(output_clip_path):
        raise FileNotFoundError("The clip wasn't extracted")

    else:
        os.chmod(output_clip_path, 0o777)
        logging.info("Clip extracted successfully")


def create_clips(
    available_movies_df: pd.DataFrame,
    selected_movies: str,
    movies_paths: str,
    clip_selection,
    project: Project,
    modification_details: dict,
    gpu_available: bool,
    is_example: bool,
    pool_size: int = 4,
):
    """
    This function takes a movie and extracts clips from it

    :param available_movies_df: the dataframe with the movies that are available for the project
    :param selected_movies: the name(s) of the movie(s) you want to extract clips from
    :param movies_paths: the path(s) to the movie(s) you want to extract clips from
    :param clip_selection: a ClipSelection object
    :param project: the project object
    :param modification_details: a dictionary with the following keys:
    :param gpu_available: True or False, depending on whether you have a GPU available to use
    :param is_example: If True and there are clip modifications, two clips will be generated (original and modified).
    :param pool_size: the number of threads to use to extract the clips, defaults to 4 (optional)
    :return: A dataframe with the clip_path, clip_filename, clip_length, upl_seconds, and clip_modification_details
    """

    # Store the desired length of the clips
    clip_length = clip_selection.kwargs["clip_length"]

    # Store the starting seconds of the clips
    if isinstance(clip_selection.result, int):
        # Clipping video from a range of seconds (e.g. 10-180)
        # Store the starting and ending of the range
        start_trim = clip_selection.kwargs["clips_range"][0]
        end_trim = clip_selection.kwargs["clips_range"][1]

        # Create a list with the starting seconds of the clips
        list_clip_start = [
            list(
                range(
                    start_trim,
                    start_trim
                    + math.floor((end_trim - start_trim) / clip_length) * clip_length,
                    clip_length,
                )
            )
        ]

        if not clip_selection.result == len(list_clip_start[0]):
            if clip_selection.result < len(list_clip_start[0]):
                # Choose random starting points based on the number of samples (example clips, by default this is 3)
                list_clip_start = [
                    np.random.choice(
                        list_clip_start[0], size=clip_selection.result, replace=False
                    )
                ]
            else:
                logging.info(
                    "There was an issue estimating the starting seconds for the clips."
                )

    else:
        # Clipping specific sections of a video at random (e.g. starting at 10, 20, 180)
        # Store the starting seconds of the clips
        list_clip_start = [clip_selection.result["clip_start_time"]]

    # Filter the df for the movie of interest
    selected_movies_df = available_movies_df[
        available_movies_df["filename"] == selected_movies
    ].reset_index(drop=True)

    # Add the list of starting seconds to the df
    selected_movies_df["list_clip_start"] = list_clip_start

    # Use explode() to expand the list column
    clips_start_df = selected_movies_df.explode("list_clip_start")

    # Rename the exploded column
    clips_start_df.rename(columns={"list_clip_start": "upl_seconds"}, inplace=True)

    # Add the length of the clips to df (to keep track of the length of each uploaded clip)
    clips_start_df["clip_length"] = clip_length

    # Specify output path for zooniverse clip extraction
    if project.server == "SNIC":
        if Path("/mimer").exists():
            temp_path = "/mimer/NOBACKUP/groups/snic2021-6-9/tmp_dir"
        elif Path("/tmp").exists() and not Path("/mimer").exists():
            temp_path = "/tmp"
        else:
            logging.error("No suitable writable path found.")
            return
    else:
        temp_path = "."
    clips_folder = str(Path(temp_path, selected_movies + "_zooniverseclips"))

    # Set the filename of the clips
    clips_start_df["clip_filename"] = clips_start_df.apply(
        lambda row: f"{selected_movies}_clip_{row['upl_seconds']}_{clip_length}.mp4",
        axis=1,
    )

    # Set the path of the clips
    clips_start_df["clip_path"] = clips_start_df["clip_filename"].apply(
        lambda x: str(Path(clips_folder, x))
    )

    # Create the folder to store the videos if not exist
    if os.path.exists(clips_folder):
        shutil.rmtree(clips_folder)
    Path(clips_folder).mkdir(parents=True, exist_ok=True)
    # Recursively add permissions to folders created
    [os.chmod(root, 0o777) for root, dirs, files in os.walk(clips_folder)]

    # Add information on the modification of the clips
    clips_start_df["clip_modification_details"] = str(modification_details)

    if is_example and modification_details.children[0].value > 0:
        # Set the filename of the original clips to be compared against the modified
        clips_start_df["clip_example_original_filename"] = clips_start_df.apply(
            lambda row: f"{selected_movies}_clip_example_original_{row['upl_seconds']}_{clip_length}.mp4",
            axis=1,
        )

        # Set the path of the clips
        clips_start_df["clip_example_original_path"] = clips_start_df[
            "clip_example_original_filename"
        ].apply(lambda x: str(Path(clips_folder, x)))

        logging.info("Extracting original clips")
        # Read each movie and extract the original clips
        for index, row in tqdm(
            clips_start_df.iterrows(), total=clips_start_df.shape[0]
        ):
            # Extract the videos and store them in the folder
            extract_clips(
                movie_path=movies_paths,
                clip_length=clip_length,
                upl_second_i=row["upl_seconds"],
                output_clip_path=row["clip_example_original_path"],
                modification_details={},
                gpu_available=gpu_available,
            )
        logging.info("Original clips extracted")

    logging.info("Extracting clips")
    # Read each movie and extract the clips from the original videos
    for index, row in tqdm(clips_start_df.iterrows(), total=clips_start_df.shape[0]):
        extract_clips(
            movie_path=movies_paths,
            clip_length=clip_length,
            upl_second_i=row["upl_seconds"],
            output_clip_path=row["clip_path"],
            modification_details=modification_details,
            gpu_available=gpu_available,
        )
    logging.info("Clips successfully extracted")

    return clips_start_df


def check_clip_size(clips_list: list):
    """
    > This function takes a list of file paths and returns a dataframe with the file path and size of
    each file. If the size is too large, we suggest compressing them as a first step.

    :param clips_list: list of file paths to the clips you want to check
    :type clips_list: list
    :return: A dataframe with the file path and size of each clip
    """

    # Get list of files with size
    if clips_list is None:
        logging.error("No clips found.")
        return None
    files_with_size = [
        (file_path, Path(file_path).stat().st_size / float(1 << 20))
        for file_path in clips_list
    ]

    df = pd.DataFrame(files_with_size, columns=["File_path", "Size"])

    if df["Size"].ge(8).any():
        logging.info(
            "Clips are too large (over 8 MB) to be uploaded to Zooniverse. Compress them!"
        )
        return df
    else:
        logging.info(
            "Clips are a good size (below 8 MB). Ready to be uploaded to Zooniverse"
        )
        return df


def set_zoo_clip_metadata(
    project: Project,
    generated_clipsdf: pd.DataFrame,
    sitesdf: pd.DataFrame,
    moviesdf: pd.DataFrame,
):
    """
    This function updates the dataframe of clips to be uploaded with
    metadata about the site and project

    :param project: the project object
    :param generated_clipsdf: a df with the information of the clips to be uploaded
    :param sitesdf: a df with the information of the sites of the project
    :param moviesdf: a df with the information of the movies of the project
    :return: upload_to_zoo, sitename, created_on

    """

    # Add spyfish-specific info
    if project.Project_name in ["Spyfish_Aotearoa", "Spyfish_BOPRC"]:
        # Rename the site columns to match standard cols names
        sitesdf = sitesdf.rename(
            columns={
                "schema_site_id": "id",
                "SiteName": "old_siteName",
                "SiteID": "#siteName",
            }
        )
        generated_clipsdf = generated_clipsdf.rename(columns={"siteName": "#siteName"})

        upload_to_zoo = generated_clipsdf.merge(sitesdf, on="#siteName")

    # Combine site info to the generated_clips df
    elif "site_id" in generated_clipsdf.columns:
        # Rename the id column to match generated_clipsdf
        sitesdf = sitesdf.rename(columns={"id": "site_id", "siteName": "#siteName"})

        upload_to_zoo = generated_clipsdf.merge(sitesdf, on="site_id")
    else:
        raise ValueError("Sites table empty. Perhaps try to rebuild the initial db.")

    sitename = upload_to_zoo["#siteName"].unique()[0]

    # Rename columns to match schema names
    # (fields that begin with “#” or “//” will never be shown to volunteers)
    # (fields that begin with "!" will only be available for volunteers on the Talk section, after classification)
    upload_to_zoo = upload_to_zoo.rename(
        columns={
            "id": "movie_id",
            "created_on": "#created_on",
            "clip_length": "#clip_length",
            "filename": "#VideoFilename",
            "clip_modification_details": "#clip_modification_details",
        }
    )

    # Convert datetime to string to avoid JSON seriazible issues
    upload_to_zoo["#created_on"] = upload_to_zoo["#created_on"].astype(str)
    created_on = upload_to_zoo["#created_on"].unique()[0]

    # Select only relevant columns
    upload_to_zoo = upload_to_zoo[
        [
            "movie_id",
            "clip_path",
            "upl_seconds",
            "#clip_length",
            "#created_on",
            "#VideoFilename",
            "#siteName",
            "#clip_modification_details",
        ]
    ]

    # Add information about the type of subject
    upload_to_zoo["Subject_type"] = "clip"

    # Add spyfish-specific info
    if project.Project_name == "Spyfish_Aotearoa":
        # Rename columns to match schema names
        sitesdf = sitesdf.rename(
            columns={
                "LinkToMarineReserve": "!LinkToMarineReserve",
            }
        )

        # Select only relevant columns
        sitesdf = sitesdf[["!LinkToMarineReserve", "#siteName", "ProtectionStatus"]]

        # Include site info to the df
        upload_to_zoo = upload_to_zoo.merge(sitesdf, on="#siteName")

    if project.Project_name == "Koster_Seafloor_Obs":
        # Rename columns to match schema names
        sitesdf = sitesdf.rename(
            columns={
                "decimalLatitude": "#decimalLatitude",
                "decimalLongitude": "#decimalLongitude",
                "geodeticDatum": "#geodeticDatum",
                "countryCode": "#countryCode",
            }
        )

        # Select only relevant columns
        sitesdf = sitesdf[
            [
                "#siteName",
                "#decimalLatitude",
                "#decimalLongitude",
                "#geodeticDatum",
                "#countryCode",
            ]
        ]

        # Include site info to the df
        upload_to_zoo = upload_to_zoo.merge(
            sitesdf, left_on="#siteName", right_on="#siteName"
        )

    # Prevent NANs on any column
    if upload_to_zoo.isnull().values.any():
        logging.info(
            f"The following columns have NAN values {upload_to_zoo.columns[upload_to_zoo.isna().any()].tolist()}"
        )

    logging.info(f"The metadata for the {upload_to_zoo.shape[0]} subjects is ready.")

    return upload_to_zoo, sitename, created_on


def upload_clips_to_zooniverse(
    project: Project,
    upload_to_zoo: pd.DataFrame,
    sitename: str,
    created_on: str,
):
    """
    It takes a dataframe of clips and metadata, creates a new subject set, and uploads the clips to
    Zooniverse

    :param upload_to_zoo: the dataframe of clips to upload
    :param sitename: the name of the site you're uploading clips from
    :param created_on: the date the clips were created
    :param project: the project object
    """

    # Estimate the number of clips
    n_clips = upload_to_zoo.shape[0]

    # Create a new subject set to host the clips
    subject_set = SubjectSet()
    subject_set_name = "clips_" + sitename + "_" + str(int(n_clips)) + "_" + created_on
    subject_set.links.project = project.Zooniverse_number
    subject_set.display_name = subject_set_name
    subject_set.save()

    logging.info(f"{subject_set_name} subject set created")

    # Save the df as the subject metadata
    subject_metadata = upload_to_zoo.set_index("clip_path").to_dict("index")

    # Upload the clips to Zooniverse (with metadata)
    new_subjects = []

    logging.info("Uploading subjects to Zooniverse")
    for clip_path, metadata in tqdm(
        subject_metadata.items(), total=len(subject_metadata)
    ):
        # Create a subject
        subject = Subject()

        # Add project info
        subject.links.project = project.Zooniverse_number

        # Add location of clip
        subject.add_location(clip_path)

        # Add metadata
        subject.metadata.update(metadata)

        # Save subject info
        subject.save()
        new_subjects.append(subject)

    # Upload all subjects
    subject_set.add(new_subjects)

    logging.info("Subjects uploaded to Zooniverse")


##########################
# Upload frames functions
##########################
def extract_frames_for_zoo(
    project: Project,
    species_list: list,
    zoo_info: dict,
    agg_df: pd.DataFrame,
    db_connection,
    server_connection,
    n_frames_subject,
    subsample_up_to,
):
    """
    > This function allows you to choose a species of interest, and then it will fetch a random
    sample of frames from the database for that species

    :param n_frames_subject: number of frames to fetch per subject, defaults to 3
    :type n_frames_subject: int (optional)
    :param subsample_up_to: If you have a lot of frames for a given species, you can subsample them.
           This parameter controls how many subjects you want to subsample to, defaults to 100
    :type subsample_up_to: int (optional)
    """

    # Roadblock to check if species list is empty
    if len(species_list) == 0:
        raise ValueError(
            "No species were selected. Please select at least one species before continuing."
        )

    # Select only aggregated classifications of species of interest
    sp_agg_df = agg_df[agg_df["label"].isin(species_list)].reset_index(drop=True)

    # Subsample up to n subjects per label
    if sp_agg_df["label"].value_counts().max() > subsample_up_to:
        logging.info(
            f"Subsampling up to {subsample_up_to} subjects of the species selected"
        )
        # randomise the order and select up to n per group
        sp_agg_df = sp_agg_df.sample(frac=1).groupby("label").head(subsample_up_to)

    # Identify the second of the original movie when the species first appears
    sp_agg_df["first_seen_movie"] = (
        sp_agg_df["clip_start_time"] + sp_agg_df["first_seen"]
    )

    # Add information of the original movies associated with the subjects
    # (e.g. the movie that was clipped from)
    (
        movies_df,
        no_available_movies_df,
        no_info_movies_df,
    ) = movie_utils.retrieve_movie_info_from_server(
        project=project,
        server_connection=server_connection,
        db_connection=db_connection,
    )

    # Include movies' filepath and fps to the df
    comb_df = sp_agg_df.merge(movies_df, on="movie_id", suffixes=("", "_df2"))

    # Prevent trying to extract frames from movies that are not accessible
    if len(comb_df[~comb_df.exists]) > 0:
        logging.error(
            f"There are {len(comb_df) - comb_df.exists.sum()} out of"
            "{len(frames_df)} subjects with original movies that are not accessible"
        )

    # Get csv paths
    csv_paths = server_utils.download_init_csv(
        project, ["movies", "species", "photos", "surveys", "sites"], server_connection
    )

    # Create a commonName col based on the zoo label to match schema format
    comb_df["commonName"] = comb_df["label"]

    # Combine the aggregated clips and species dataframes
    comb_df = db_utils.add_db_info_to_df(
        project=project,
        conn=db_connection,
        df=comb_df,
        csv_paths=csv_paths,
        table_name="species",
        cols_interest="id, commonName, scientificName",
    )

    # Create a list with the frames to be extracted and save into frame_number column
    comb_df["frame_number"] = comb_df[["first_seen_movie", "fps"]].apply(
        lambda x: [
            int((x["first_seen_movie"] + j) * x["fps"]) for j in range(n_frames_subject)
        ],
        1,
    )

    # Reshape df to have each frame as rows
    lst_col = "frame_number"

    comb_df = pd.DataFrame(
        {
            col: np.repeat(comb_df[col].values, comb_df[lst_col].str.len())
            for col in comb_df.columns.difference([lst_col])
        }
    ).assign(**{lst_col: np.concatenate(comb_df[lst_col].values)})[
        comb_df.columns.tolist()
    ]

    # Drop unnecessary columns
    comb_df.drop(["subject_ids"], inplace=True, axis=1)

    # Check the frames haven't been uploaded to Zooniverse
    comb_df = check_frames_uploaded(db_connection, comb_df)

    # Specify the temp location to store the frames
    temp_frames_folder = "_".join(species_list) + "_frames/"
    # Make sure that folder names do not exceed a certain length
    if len(temp_frames_folder) > 260:
        curr = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        temp_frames_folder = f"{curr}_various_frames/"

    if project.server == "SNIC":
        # Specify volume allocated by SNIC
        if Path("/mimer").exists():
            folder_name = "/mimer/NOBACKUP/groups/snic2021-6-9/tmp_dir/frames"
        elif Path("/tmp").exists() and not Path("/mimer").exists():
            folder_name = "/tmp/frames"
        else:
            logging.error("No suitable writable path found.")
            return
        frames_folder = str(Path(folder_name, temp_frames_folder))
    else:
        frames_folder = temp_frames_folder
    #         if hasattr(project, "output_path"):
    #             mod_frames_folder = project.output_path + frames_folder

    # Extract the frames from the videos, store them in the temp location
    # and save the df with information about the frames in the projectprocessor
    generated_frames = movie_utils.extract_frames(
        project=project,
        server_connection=server_connection,
        df=comb_df,
        frames_folder=frames_folder,
    )
    return generated_frames


def check_frame_size(frame_paths: list):
    """
    It takes a list of file paths, gets the size of each file, and returns a dataframe with the file
    path and size of each file

    :param frame_paths: a list of paths to the frames you want to check
    :return: A dataframe with the file path and size of each frame.
    """

    # Get list of files with size
    files_with_size = [
        (file_path, os.stat(file_path).st_size) for file_path in frame_paths
    ]

    df = pd.DataFrame(files_with_size, columns=["File_path", "Size"])

    # Change bytes to MB
    df["Size"] = df["Size"] / 1000000

    if df["Size"].ge(1).any():
        logging.info(
            "Frames are too large (over 1 MB) to be uploaded to Zooniverse. Compress them!"
        )
        return df
    else:
        logging.info(
            "Frames are a good size (below 1 MB). Ready to be uploaded to Zooniverse"
        )
        return df


# Function to gather information of frames already uploaded to Zooniverse
def check_frames_uploaded(
    db_connection,
    frames_df: pd.DataFrame,
):
    from kso_utils.db_utils import get_df_from_db_table

    # Get info of the subjects uploaded to zooniverse from the db
    subjects_df = get_df_from_db_table(db_connection, "subjects")

    # Select only frame subjects
    subjects_df = subjects_df[subjects_df["subject_type"] == "frame"]

    if subjects_df.empty:
        return frames_df

    # Create a list of the species about to upload (scientificName)
    list_species = frames_df["scientificName"].unique()

    # Query id and sci. names from the species table
    species_df = get_df_from_db_table(db_connection, "species")[
        ["id", "scientificName"]
    ]

    # Rename columns to match subject df
    species_df = species_df.rename(columns={"id": "frame_exp_sp_id"})

    # Reference the expected species on the uploaded subjects
    subjects_df = pd.merge(
        subjects_df,
        species_df,
        how="left",
        on="frame_exp_sp_id",
    )

    # Select only those subjects that should have the species of interest
    uploaded_frames_df = subjects_df[subjects_df["scientificName"].isin(list_species)]

    uploaded_frames_df = uploaded_frames_df[
        ["movie_id", "frame_number", "frame_exp_sp_id"]
    ]

    # Filter out frames that have already been uploaded
    if (
        len(uploaded_frames_df) > 0
        and not uploaded_frames_df["frame_number"].isnull().any()
    ):
        logging.info(
            "There are some frames already uploaded in Zooniverse for the species selected. \
            Checking if those are the frames you are trying to upload"
        )
        # Ensure that frame_number is an integer
        uploaded_frames_df["frame_number"] = uploaded_frames_df["frame_number"].astype(
            int
        )
        frames_df["frame_number"] = frames_df["frame_number"].astype(int)
        merge_df = (
            pd.merge(
                frames_df,
                uploaded_frames_df,
                left_on=["movie_id", "frame_number"],
                right_on=["movie_id", "frame_number"],
                how="left",
                indicator=True,
            )["_merge"]
            == "both"
        )

        # Exclude frames that have already been uploaded
        # trunk-ignore(flake8/E712)
        frames_df = frames_df[~merge_df]
        if len(frames_df) == 0:
            logging.error("All of the frames you have selected are already uploaded.")
        else:
            logging.info(
                f"There are {len(frames_df)} frames with the species of interest not uploaded to Zooniverse yet.",
            )

    else:
        logging.info(
            "There are no frames uploaded in Zooniverse for the species selected."
        )

    return frames_df


# Function modify the frames
def modify_frames(
    project: Project,
    frames_to_upload_df: pd.DataFrame,
    species_i: list,
    modification_details: dict,
):
    if len(species_i) == 0:
        species_i = ["custom_species"]

    # Specify the folder to host the modified frames
    if project.server == "SNIC":
        # Specify volume allocated by SNIC
        if Path("/mimer").exists():
            folder_name = "/mimer/NOBACKUP/groups/snic2021-6-9/tmp_dir/frames"
        elif Path("/tmp").exists() and not Path("/mimer").exists():
            folder_name = "/tmp/frames"
        else:
            logging.error("No suitable writable path found.")
            return
        mod_frames_folder = Path(
            folder_name, "modified_" + "_".join(species_i) + "_frames"
        )
    else:
        mod_frames_folder = Path("modified_" + "_".join(species_i) + "_frames")
        if hasattr(project, "output_path"):
            mod_frames_folder = project.output_path / mod_frames_folder

    # Specify the path of the modified frames
    frames_to_upload_df["modif_frame_path"] = frames_to_upload_df["frame_path"].apply(
        lambda x: str(mod_frames_folder / Path(x).name)
    )

    # Remove existing modified clips
    if mod_frames_folder.exists():
        shutil.rmtree(mod_frames_folder)

    if len(modification_details.values()) > 0:
        # Save the modification details to include as subject metadata
        frames_to_upload_df["frame_modification_details"] = str(modification_details)

        # Create the folder to store the videos if not exist
        if not mod_frames_folder.exists():
            mod_frames_folder.mkdir(parents=True, exist_ok=True)
            # Recursively add permissions to folders created
            [os.chmod(root, 0o777) for root, _, _ in os.walk(mod_frames_folder)]

        # Modify the clips###
        # Read each clip and modify them (showing a progress bar)
        for index, row in tqdm(
            frames_to_upload_df.iterrows(), total=frames_to_upload_df.shape[0]
        ):
            if not row["modif_frame_path"].exists():
                # Set up input prompt
                init_prompt = f'ffmpeg.input("{row["frame_path"]}")'
                full_prompt = init_prompt
                # Set up modification
                for transform in modification_details.values():
                    if "filter" in transform:
                        mod_prompt = transform["filter"]
                        full_prompt += mod_prompt

                # Run the modification
                try:
                    logging.info(full_prompt)
                    eval(full_prompt).run(capture_stdout=True, capture_stderr=True)
                    row["modif_frame_path"].chmod(0o777)
                except ffmpeg.Error as e:
                    logging.info("stdout: {}", e.stdout.decode("utf8"))
                    logging.info("stderr: {}", e.stderr.decode("utf8"))
                    raise e

        logging.info("Frames modified successfully")

    else:
        # Save the modification details to include as subject metadata
        frames_to_upload_df["modif_frame_path"] = frames_to_upload_df["frame_path"]

    return frames_to_upload_df


# Function to set the metadata of the frames to be uploaded to Zooniverse
def set_zoo_frame_metadata(
    project: Project,
    db_connection,
    df: pd.DataFrame,
    species_list: list,
    csv_paths: dict,
):
    """
    It takes a dataframe of clips or frames, and adds metadata about the site and project to it

    :param df: the dataframe with the media to upload
    :param project: the project object
    :param species_list: a list of the species that should be on the frames
    :param csv_paths: a dictionary with the paths of the csvs used to initiate the db
    :return: upload_to_zoo, sitename, created_on
    """
    project_name = project.Project_name

    if not isinstance(df, pd.DataFrame):
        df = df.df

    if (
        "modif_frame_path" in df.columns
        and "no_modification" not in df["modif_frame_path"].values
    ):
        df["frame_path"] = df["modif_frame_path"].apply(str)

    # Roadblock to prevent uploading frames to template project:
    if project.Zooniverse_number == 9754:
        raise ValueError(
            "You are not allowed to upload frames to the template Zooniverse project."
        )

    # Set project-specific metadata
    if project.Zooniverse_number == 9747:
        df = add_db_info_to_df(
            project, db_connection, csv_paths, df, "sites", "id, siteName"
        )
        upload_to_zoo = df[
            [
                "frame_path",
                "frame_number",
                "species_id",
                "movie_id",
                "created_on",
                "siteName",
            ]
        ]

    elif project_name == "SGU":
        upload_to_zoo = df[["frame_path", "species_id", "filename"]]

    elif project_name == "Spyfish_Aotearoa":
        from kso_utils.spyfish_utils import spyfish_subject_metadata

        upload_to_zoo = spyfish_subject_metadata(df, csv_paths=csv_paths)
    else:
        logging.error("This project is not a supported Zooniverse project.")

    # Add information about the type of subject
    upload_to_zoo = upload_to_zoo.copy()
    upload_to_zoo.loc[:, "subject_type"] = "frame"
    upload_to_zoo = upload_to_zoo.rename(columns={"species_id": "frame_exp_sp_id"})

    # Check there are no empty values (prevent issues uploading subjects)
    if upload_to_zoo.isnull().values.any():
        logging.error(
            "There are some values missing from the data you are trying to upload."
        )

    return upload_to_zoo


# Function to upload frames to Zooniverse
def upload_frames_to_zooniverse(
    project: Project,
    upload_to_zoo: pd.DataFrame,
    species_list: list,
):
    """
    It takes a dataframe of frames, and upload it to Zooniverse

    :param df: the dataframe with the media to upload
    :param project: the project object
    :param species_list: a list of the species that should be on the frames
    :return: upload_to_zoo, sitename, created_on
    """

    # Retireve zooniverse project name and number
    project_name = project.Project_name
    project_number = project.Zooniverse_number

    # Estimate the number of frames
    n_frames = upload_to_zoo.shape[0]

    if project_name == "Koster_Seafloor_Obs":
        created_on = upload_to_zoo["created_on"].unique()[0]
        sitename = upload_to_zoo["siteName"].unique()[0]

        # Name the subject set
        subject_set_name = (
            "frames_"
            + str(int(n_frames))
            + "_"
            + "_".join(species_list)
            + "_"
            + sitename
            + "_"
            + created_on
        )

    elif project_name == "SGU":
        surveys_df = pd.read_csv(project.csv_paths["local_surveys_csv"])
        created_on = surveys_df["SurveyDate"].unique()[0]
        folder_name = Path(upload_to_zoo["frame_path"].iloc[0]).parent.name
        sitename = folder_name

        # Name the subject set
        subject_set_name = (
            "frames_"
            + str(int(n_frames))
            + "_"
            + "_".join(species_list)
            + "_"
            + sitename
            + "_"
            + created_on
        )

    elif project_name == "Spyfish_Aotearoa":
        reserve = upload_to_zoo["!LinkToMarineReserve"].unique()[0]

        # Name the subject for frames from multiple sites/movies
        subject_set_name = (
            "frames_"
            + str(int(n_frames))
            + "_"
            + reserve
            + "_"
            + "_".join(species_list)
            + datetime.date.today().strftime("_%d_%m_%Y")
        )

    else:
        sitename = upload_to_zoo["siteName"].unique()[0]

        # Name the subject for frames from multiple sites/movies
        subject_set_name = (
            "frames_"
            + str(int(n_frames))
            + "_"
            + sitename
            + "_"
            + "_".join(species_list)
            + datetime.date.today().strftime("_%d_%m_%Y")
        )

    # Create a new subject set to host the frames
    subject_set = SubjectSet()
    subject_set.links.project = project_number
    subject_set.display_name = subject_set_name
    subject_set.save()

    logging.info(f"{subject_set_name} subject set created")

    # Save the df as the subject metadata
    subject_metadata = upload_to_zoo.set_index("frame_path").to_dict("index")

    # Upload the clips to Zooniverse (with metadata)
    new_subjects = []

    logging.info("Uploading subjects to Zooniverse...")
    for frame_path, metadata in tqdm(
        subject_metadata.items(), total=len(subject_metadata)
    ):
        subject = Subject()

        subject.links.project = project_number
        subject.add_location(frame_path)

        logging.info(frame_path)
        subject.metadata.update(metadata)

        logging.info(metadata)
        subject.save()
        logging.info("Subject saved")
        new_subjects.append(subject)

    # Upload videos
    subject_set.add(new_subjects)
    logging.info("Subjects uploaded to Zooniverse")



import requests
import os
import json
import zipfile
import tarfile
import logging
from pathlib import Path

# Logging
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


def zip_folder(folder_path):
    folder_path = Path(folder_path)
    zip_file_name = folder_path.with_suffix(".zip")

    with zipfile.ZipFile(zip_file_name, "w", zipfile.ZIP_DEFLATED) as zipf:
        for file_path in folder_path.glob("**/*"):
            if file_path.is_file():
                zipf.write(file_path, file_path.relative_to(folder_path))

    return zip_file_name


def upload_archive(access_key: str, artifact_dir: str):
    """
    > Uploads the last file in the `artifact_dir` to Zenodo

    :param access_key: the access token you got from Zenodo
    :type access_key: str
    :param bucket_url: the url of the bucket you want to upload to
    :type bucket_url: str
    :param file_path: the path to the file you want to upload
    :type file_path: str
    :param artifact_dir: the directory where the artifacts are stored
    :type artifact_dir: str
    """

    depo_id, bucket_url = get_zenodo_id_bucket(access_key=access_key)

    add_file_to_zenodo_upload(
        access_key,
        bucket_url,
        artifact_dir,
    )

    return depo_id


def download_and_extract_models_from_zenodo(api_token, download_dir="models"):
    """
    Download model archives from Zenodo associated with the account, extract the model files, and return their paths.

    Parameters:
    api_token (str): Zenodo API token.
    download_dir (str): Directory where models will be downloaded and extracted.

    Returns:
    dict: Dictionary with model names as keys and local file paths as values.
    """
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    headers = {"Authorization": f"Bearer {api_token}"}

    model_paths = {}
    page = 1
    base_url = "https://zenodo.org/api/deposit/depositions"

    while True:
        response = requests.get(base_url, headers=headers, params={"page": page})
        if response.status_code != 200:
            print(
                f"Failed to retrieve records. HTTP status code: {response.status_code}"
            )
            break

        records = response.json()
        if not records:
            break

        for record in records:
            for file_info in record["files"]:
                file_name = file_info["filename"]
                if "ref" in file_name:
                    download_url = file_info["links"]["download"]
                    download_url = download_url.replace("/draft", "")
                    local_filename = os.path.join(download_dir, file_name)

                    print(f"Downloading {file_name} from {download_url}...")
                    file_response = requests.get(
                        download_url, headers=headers, stream=True
                    )
                    if file_response.status_code == 200:
                        with open(local_filename, "wb") as f:
                            for chunk in file_response.iter_content(chunk_size=8192):
                                f.write(chunk)
                        print(f"{file_name} downloaded successfully.")

                        # Extract the archive
                        extracted_files = extract_archive(local_filename, download_dir)
                        for extracted_file in extracted_files:
                            if extracted_file.endswith(".pt"):
                                model_paths[
                                    os.path.basename(local_filename).replace(".zip", "")
                                ] = os.path.join(download_dir, extracted_file)

                        # Remove the archive after extraction
                        os.remove(local_filename)
                    else:
                        print(
                            f"Failed to download {file_name}. HTTP status code: {file_response.status_code}"
                        )

        page += 1

    return model_paths


def extract_archive(archive_path, extract_to):
    """
    Extract an archive file and return a list of extracted file paths.

    Parameters:
    archive_path (str): Path to the archive file.
    extract_to (str): Directory where the archive will be extracted.

    Returns:
    list: List of extracted file paths.
    """
    extracted_files = []

    if archive_path.endswith(".zip"):
        with zipfile.ZipFile(archive_path, "r") as zip_ref:
            zip_ref.extractall(extract_to)
            extracted_files = zip_ref.namelist()
    elif archive_path.endswith((".tar.gz", ".tgz")):
        with tarfile.open(archive_path, "r:gz") as tar_ref:
            tar_ref.extractall(extract_to)
            extracted_files = tar_ref.getnames()
    elif archive_path.endswith(".tar"):
        with tarfile.open(archive_path, "r:") as tar_ref:
            tar_ref.extractall(extract_to)
            extracted_files = tar_ref.getnames()
    else:
        print(f"Unsupported archive format: {archive_path}")

    return extracted_files


# Get deposition id, i.e. "id" field from this response and bucket
def get_zenodo_id_bucket(access_key: str):
    headers = {"Content-Type": "application/json"}
    params = {"access_token": access_key}
    r = requests.post(
        "https://zenodo.org/api/deposit/depositions",
        params=params,
        json={},
        # Headers are not necessary here since "requests" automatically
        # adds "Content-Type: application/json", because we're using
        # the "json=" keyword argument
        # headers=headers,
        headers=headers,
    )
    response = r.json()
    return response["id"], response["links"]["bucket"]


def add_file_to_zenodo_upload(access_key: str, bucket_url: str, file_path: str):
    filename = os.path.basename(file_path)
    # The target URL is a combination of the bucket link with the desired filename
    # seperated by a slash.
    params = {"access_token": access_key}
    with open(file_path, "rb") as fp:
        r = requests.put(
            "%s/%s" % (bucket_url, filename),
            data=fp,
            params=params,
        )
    return r.json()


def add_metadata_zenodo_upload(
    access_token: str,
    deposition_id: str,
    title: str,
    description: str,
    creators_dict: dict,
):
    # Add metadata
    data = {
        "metadata": {
            "title": title,
            "upload_type": "software",
            "description": description,
            "creators": [
                {"name": name, "affiliation": affiliation}
                for name, affiliation in creators_dict.items()
            ],
            "communities": [{"identifier": "odf-sweden"}],
            "notes": "Attribution notice: The code used to generate this model can be found "
            "at https://github.com/ocean-data-factory-sweden/koster_data_management",
        }
    }
    headers = {"Content-Type": "application/json"}
    r = requests.put(
        f"https://zenodo.org/api/deposit/depositions/{deposition_id}",
        params={"access_token": access_token},
        data=json.dumps(data),
        headers=headers,
    )
    if r.status_code == 200:
        logging.info("Upload successful")
        r = requests.post(
            f"https://zenodo.org/api/deposit/depositions/{deposition_id}/actions/publish",
            params={"access_token": access_token},
        )
        return r.status_code

    else:
        logging.info("Upload failed")
